#include "anime.h"
#include "libmy/lex.h"
#include "libmy/key.h"
#include "img.h"
#include "bgm.h"
#include "res.h"


//---------------------------------------------------------------------------
ST_ANIME_TABLE AnimePat[ANIME_MAX_PAT_CNT] = {
	{ "bg",     (void*)AnimeExecBg     },
	{ "crop",   (void*)AnimeExecCrop   },
	{ "crop2",  (void*)AnimeExecCrop2  },
	{ "crop3",  (void*)AnimeExecCrop3  },
	{ "blend",  (void*)AnimeExecBlend  },
	{ "effect", (void*)AnimeExecEffect },
	{ "wait",   (void*)AnimeExecWait   },
	{ "bgm",    (void*)AnimeExecBgm    },
	{ "var",    (void*)AnimeExecVar    },
	{ "pos",    (void*)AnimeExecPos    },
	{ "inc",    (void*)AnimeExecInc    },
	{ "dec",    (void*)AnimeExecDec    },
	{ "jae" ,   (void*)AnimeExecJae    },
	{ "jbe" ,   (void*)AnimeExecJbe    },
	{ "end",    (void*)AnimeExecEnd    },
};

char* AnimeDat[ANIME_MAX_DAT_CNT] = {
	(char*)&ani_501_txt,		// S_OP01`S_OP08
	(char*)&ani_502_txt,		// op
	NULL,
	(char*)&ani_504_txt,		// ed
	NULL,
	NULL,
	NULL,
	(char*)&ani_508_txt,		// S_OP98
	(char*)&ani_509_txt,		// S_OP99
	NULL,
	(char*)&ani_511_txt,		// S_SIO1
	(char*)&ani_512_txt,		// S_NAY1, BG004N_2
	(char*)&ani_513_txt,		// S_MAI1
	(char*)&ani_514_txt,		// S_MAI2

	// ǉ
	(char*)&ani_600_txt,		// title
	(char*)&ani_601_txt,		// SAYU_01
	(char*)&ani_602_txt,		// S_AYU1
	(char*)&ani_603_txt,		// S_AYU2
	(char*)&ani_604_txt,		// S_SIO2
};

//---------------------------------------------------------------------------
ST_ANIME Anime;


//---------------------------------------------------------------------------
EWRAM_CODE void AnimeInit(void)
{
	_Memset(&Anime, 0x00, sizeof(ST_ANIME));

	Anime.act = ANIME_ACT_END;
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeSetDat(u32 no)
{
	TRACEOUT("[AnimeSetDat %d]\n", no);
	_ASSERT(no < ANIME_MAX_DAT_CNT);


	LexSetCur(LEX_ID_ANIME, AnimeDat[no]);

	AnimeInit();
	Anime.no = no;
	Anime.act = ANIME_ACT_EXEC;
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExec(void)
{
	if(Anime.act == ANIME_ACT_END || ImgIsEnd() == FALSE)
	{
		return;
	}

	// ^CgȊÕXLbv
	if(Anime.no != 14 && KeyIsCntAB() == TRUE)
	{
		ImgSetBg("SIRO");
		ImgSetEffect(IMG_EFFECT_0);
		BgmStop();

		Anime.act = ANIME_ACT_END;
		return;
	}

	if(Anime.wait != 0)
	{
		Anime.wait--;
		return;
	}

	AnimeExecSub();
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecSub(void)
{
	char* p = LexGetStr(LEX_ID_ANIME);
	TRACEOUT("[Anime %s]\n", p);


	u32 i;

	for(i=0; i<ANIME_MAX_PAT_CNT; i++)
	{
		if(_Strncmp(p, AnimePat[i].pStr, ANIME_MAX_STR_SIZE) == 0)
		{
			break;
		}
	}


	_ASSERT(i < ANIME_MAX_PAT_CNT);
	AnimePat[i].pFunc();
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecBg(void)
{
	char* p = LexGetStr(LEX_ID_ANIME);

	ImgDrawBg(p);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecCrop(void)
{
	char* p = LexGetStr(LEX_ID_ANIME);
	u32  sx = LexGetNum(LEX_ID_ANIME);
	u32  sy = LexGetNum(LEX_ID_ANIME);

	ImgDrawCrop(sx, sy, p);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecCrop2(void)
{
	u32 sx = LexGetNum(LEX_ID_ANIME);
	u32 sy = LexGetNum(LEX_ID_ANIME);

	char buf[32];
	_Sprintf(buf, "Z%03d", Anime.var);

	ImgDrawCrop(sx, sy, buf);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecCrop3(void)
{
	char buf[32];
	_Sprintf(buf, "Z%03d", Anime.var);

	ImgDrawCropCenter(buf);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecBlend(void)
{
	char* p = LexGetStr(LEX_ID_ANIME);
	u32  sx = LexGetNum(LEX_ID_ANIME);
	u32  sy = LexGetNum(LEX_ID_ANIME);

	ImgDrawBlend(sx, sy, p);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecEffect(void)
{
	u32 v = LexGetNum(LEX_ID_ANIME);

	ImgSetEffect(v);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecWait(void)
{
	Anime.wait = LexGetNum(LEX_ID_ANIME);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecBgm(void)
{
	u32  no = LexGetNum(LEX_ID_ANIME);
	bool is = (LexGetNum(LEX_ID_ANIME) == 1) ? TRUE : FALSE;

	BgmPlay(no, is);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecVar(void)
{
	Anime.var = LexGetNum(LEX_ID_ANIME);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecPos(void)
{
	Anime.pCur = LexGetCur(LEX_ID_ANIME);
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecInc(void)
{
	Anime.var++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecDec(void)
{
	Anime.var--;
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecJae(void)
{
	u32 v = LexGetNum(LEX_ID_ANIME);

	if(Anime.var >= v)
	{
		LexSetCur(LEX_ID_ANIME, Anime.pCur);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecJbe(void)
{
	u32 v = LexGetNum(LEX_ID_ANIME);

	if(Anime.var <= v)
	{
		LexSetCur(LEX_ID_ANIME, Anime.pCur);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void AnimeExecEnd(void)
{
	Anime.act = ANIME_ACT_END;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool AnimeIsEnd(void)
{
	return (Anime.act == ANIME_ACT_END) ? TRUE : FALSE;
}
