#include "img.h"
#include "libmy/mode3.h"
#include "file.h"

//---------------------------------------------------------------------------
ST_IMG Img;


//---------------------------------------------------------------------------
EWRAM_CODE void ImgInit(void)
{
	_Memset(&Img, 0x00, sizeof(ST_IMG));
}
//---------------------------------------------------------------------------
IWRAM_CODE void ImgExec(void)
{
	if(Img.isDraw == FALSE)
	{
		return;
	}


	switch(Img.effect)
	{
	case IMG_EFFECT_0:
	case IMG_EFFECT_26:			// p
	case IMG_EFFECT_60:
	case IMG_EFFECT_105:		// p
		ImgExecEffect0();
		break;

	case IMG_EFFECT_48:
		ImgExecEffect48();
		break;

	case IMG_EFFECT_200:
	default:
		ImgExecEffect200();
		break;

	case IMG_EFFECT_NORMAL:
		ImgExecEffectNormal();
		break;

	case IMG_EFFECT_MASK:
	case IMG_EFFECT_BLEND:		// p
		ImgExecEffectMask();
		break;

	case IMG_EFFECT_LR:
		ImgExecEffectLr();
		break;

	case IMG_EFFECT_RL:
		ImgExecEffectRl();
		break;

	case IMG_EFFECT_UD:
		ImgExecEffectUd();
		break;

	case IMG_EFFECT_STR:
		ImgExecEffectStr();
		break;

	case IMG_EFFECT_OP1:
		ImgExecEffectOp1();
		break;

	case IMG_EFFECT_OP2:
		ImgExecEffectOp2();
		break;
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void ImgExecBuf(void)
{
	if(Img.bgStr[0] != '\0')
	{
		ImgDrawBg(Img.bgStr);
	}

	if(Img.bg2Str[0] != '\0')
	{
		ImgDrawBlend(0, 0, Img.bg2Str);
	}

	if(Img.chrStr[0] != '\0')
	{
		ImgDrawBlendChr(Img.chrStr);
	}

	if(Img.isDrawCalendar == TRUE)
	{
		ImgDrawCalendar();
	}
}
//---------------------------------------------------------------------------
// 킶
IWRAM_CODE void ImgExecEffect0(void)
{
	u16 cx[4] = { 0, 3, 0, 3 };
	u16 cy[4] = { 0, 3, 3, 0 };
	u16 x;
	u16 y;
	u16 idx = Img.step - 1;

	switch(Img.step)
	{
	case 0:
		ImgExecBuf();
		Img.step++;
		break;

	default:
		for(y=0; y<160; y+=8)
		{
			for(x=0; x<240; x+=8)
			{
				Mode3Copy(x+cx[idx], y+cy[idx], x+cx[idx]+4, y+cy[idx]+4);
			}
		}

		if(idx >= 3)
		{
			Img.isDraw = FALSE;
		}
		Img.step++;
		break;
	}
}
//---------------------------------------------------------------------------
// E
IWRAM_CODE void ImgExecEffect48(void)
{
	switch(Img.step)
	{
	case 0:
		ImgExecBuf();
		Img.step++;
		break;

	default: ; 
		u16 x = ((Img.step-1) * 8);
		Mode3Copy(x, 0, x+7, 159);
		Img.step++;

		if(Img.step >= 31)
		{
			Img.isDraw = FALSE;
		}
		break;
	}
}
//---------------------------------------------------------------------------
// ʏRs[
IWRAM_CODE void ImgExecEffect200(void)
{
	switch(Img.step)
	{
	case 0:
		ImgExecBuf();
		Img.step++;
		break;

	case 1:
		Mode3SetDraw(TRUE);
		Img.isDraw = FALSE;
		break;
	}
}
//---------------------------------------------------------------------------
// ʏRs[
IWRAM_CODE void ImgExecEffectNormal(void)
{
	Mode3SetDraw(TRUE);
	Img.isDraw = FALSE;
}
//---------------------------------------------------------------------------
// }XN
IWRAM_CODE void ImgExecEffectMask(void)
{
	u16 cx[4] = { 0, 3, 0, 3 };
	u16 cy[4] = { 0, 3, 3, 0 };
	u16 x;
	u16 y;
	u16 idx = Img.step;

	for(y=0; y<160; y+=8)
	{
		for(x=0; x<240; x+=8)
		{
			Mode3Copy(x+cx[idx], y+cy[idx], x+cx[idx]+4, y+cy[idx]+4);
		}
	}

	if(idx >= 3)
	{
		Img.isDraw = FALSE;
	}
	Img.step++;
}
//---------------------------------------------------------------------------
// と
IWRAM_CODE void ImgExecEffectUd(void)
{
	u16 y = ((Img.step) * 8);

	Mode3Copy(0, y, 239, y+7);

	Img.step++;
	if(Img.step >= 20)
	{
		Img.isDraw = FALSE;
	}
}
//---------------------------------------------------------------------------
// E
IWRAM_CODE void ImgExecEffectLr(void)
{
	u16 x = Img.step * 8;

	Mode3Copy(x, 0, x+7, 159);

	Img.step++;
	if(Img.step >= 30)
	{
		Img.isDraw = FALSE;
	}
}
//---------------------------------------------------------------------------
// E
IWRAM_CODE void ImgExecEffectRl(void)
{
	u16 x = (239 - 7) - (Img.step * 8);

	Mode3Copy(x, 0, x+7, 159);

	Img.step++;
	if(Img.step >= 30)
	{
		Img.isDraw = FALSE;
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void ImgExecEffectStr(void)
{
	if(Img.step >= 32*4*2)
	{
		Img.isDraw = FALSE;
		return;
	}

	Img.step++;
	if(Mod(Img.step, 4) != 0)
	{
		return;
	}

	u32 idx = (Img.step <= 32*4) ? 32-Img.step/4 : (Img.step/4)-32;

	Mode3DrawBlendWhite(0, 0, 240, 20, 0, 22, idx);
	Mode3Copy2(0, 20, 240, 22, 0, 50+Img.step/8);
}
//---------------------------------------------------------------------------
// OPXN[1i؁j
IWRAM_CODE void ImgExecEffectOp1(void)
{
	switch(Img.step)
	{
	case 0: ;
		ST_IMG_HEADER* p1 = (ST_IMG_HEADER*)FileGetImg("TATE_1");
		ST_IMG_HEADER* p2 = (ST_IMG_HEADER*)FileGetImg("TATE_2");

		Mode3DrawBg((u16*)(p1 + 1));
		Mode3DrawChr((u16*)(p2 + 1));
		Img.step++;
		break;

	default:
		if(Img.step > 160*3)
		{
			Img.isDraw = FALSE;
		}
		else if(Mod(Img.step, 3) == 0)
		{
			Mode3ScrollOp(Div(Img.step, 3));
		}
		Img.step++;
		break;
	}
}
//---------------------------------------------------------------------------
// OPXN[2ij
IWRAM_CODE void ImgExecEffectOp2(void)
{
	switch(Img.step)
	{
	case 0: ;
		ST_IMG_HEADER* p1 = (ST_IMG_HEADER*)FileGetImg("TATE_2");
		ST_IMG_HEADER* p2 = (ST_IMG_HEADER*)FileGetImg("TATE_3");

		Mode3DrawBg((u16*)(p1 + 1));
		Mode3DrawChr((u16*)(p2 + 1));
		Img.step++;
		break;

	default:
		if(Img.step > 160*3)
		{
			Img.isDraw = FALSE;
		}
		else if(Mod(Img.step, 3) == 0)
		{
			Mode3ScrollOp(Div(Img.step, 3));
		}
		Img.step++;
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawBg(char* pStr)
{
	ST_IMG_HEADER* p = (ST_IMG_HEADER*)FileGetImg(pStr);

	Mode3DrawBg((u16*)(p + 1));
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawCrop(u32 sx, u32 sy, char* pStr)
{
	ST_IMG_HEADER* p = (ST_IMG_HEADER*)FileGetImg(pStr);

	Mode3DrawCrop(sx, sy, p->cx, p->cy, (u16*)(p + 1));
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawCropCenter(char* pStr)
{
	ST_IMG_HEADER* p = (ST_IMG_HEADER*)FileGetImg(pStr);
	u32 sx = (SCREEN_WIDTH / 2) - (p->cx / 2);

	Mode3DrawCrop(sx, 0, p->cx, p->cy, (u16*)(p + 1));
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawBlend(u32 sx, u32 sy, char* pStr)
{
	ST_IMG_HEADER* pImg  = (ST_IMG_HEADER*)FileGetImg(pStr);
	ST_IMG_HEADER* pMask = (ST_IMG_HEADER*)FileGetImgMask(pStr);

	Mode3DrawBlend(sx, sy, pImg->cx, pImg->cy, (u16*)(pImg + 1), (u16*)(pMask + 1));
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawBlendChr(char* pStr)
{
	ST_IMG_HEADER* pImg  = (ST_IMG_HEADER*)FileGetImg(pStr);
	ST_IMG_HEADER* pMask = (ST_IMG_HEADER*)FileGetImgMask(pStr);

	u32 sx = (SCREEN_WIDTH / 2) - (pImg->cx / 2);
	u32 sy = (SCREEN_HEIGHT - pImg->cy);

	Mode3DrawBlend(sx, sy, pImg->cx, pImg->cy, (u16*)(pImg + 1), (u16*)(pMask + 1));
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgDrawCalendar(void)
{
	char buf1[16];
	char buf2[16];

	_Sprintf(buf1, "Z%03d", 340 + Img.day);
	_Sprintf(buf2, "Z%03d", 372 + Mod(Img.day+3, 7));

	ImgDrawBlend(4, 4, "Z379");		// Ŕ
	ImgDrawBlend(18, 16, buf1);		// t
	ImgDrawBlend(20, 34, buf2);		// j
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgSetBg(char* pStr)
{
	_Strncpy(Img.bgStr, pStr, IMG_MAX_STR_SIZE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgSetBg2(char* pStr)
{
	_Strncpy(Img.bg2Str, pStr, IMG_MAX_STR_SIZE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgSetChr(char* pStr)
{
	_Strncpy(Img.chrStr, pStr, IMG_MAX_STR_SIZE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgSetCalendar(u32 day)
{
	Img.day = day;
	Img.isDrawCalendar = TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgSetEffect(u32 no)
{
	_ASSERT(no < IMG_EFFECT_END);

	Img.effect = no;
	Img.step = 0;

	Img.isDraw = TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void ImgClear(void)
{
	ImgSetBg("");
	ImgSetBg2("");
	ImgSetChr("");

	Img.isDrawCalendar = FALSE;
}
//---------------------------------------------------------------------------
// fobOp
EWRAM_CODE void ImgDirect(char* pStr)
{
	ImgDrawBg(pStr);
	Mode3SetDraw(TRUE);
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ImgIsEnd(void)
{
	return (Img.isDraw == FALSE) ? TRUE : FALSE;
}
