#include "lex.h"


//---------------------------------------------------------------------------
ST_LEX Lex[LEX_MAX_CNT];


//---------------------------------------------------------------------------
EWRAM_CODE void LexInit(void)
{
	_Memset(&Lex, 0x00, sizeof(ST_LEX));
}
//---------------------------------------------------------------------------
EWRAM_CODE void LexSetCur(u32 id, char* pCur)
{
	Lex[id].pCur = pCur;
	Lex[id].pOld = pCur;
}
//---------------------------------------------------------------------------
EWRAM_CODE char* LexGetCur(u32 id)
{
	return Lex[id].pCur;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 LexGetNum(u32 id)
{
	LexGetToken(id, FALSE);

	if(Lex[id].type != LEX_TOKEN_NUM)
	{
		SystemError("[Err] LexGetNum -1");
	}

	return Lex[id].num;
}
//---------------------------------------------------------------------------
EWRAM_CODE char* LexGetStr(u32 id)
{
	LexGetToken(id, FALSE);

	if(Lex[id].type != LEX_TOKEN_STR)
	{
		SystemError("[Err] LexGetStr NULL");
	}

	return Lex[id].str;
}
//---------------------------------------------------------------------------
EWRAM_CODE char LexGetChr(u32 id)
{
	LexGetToken(id, TRUE);

	if(Lex[id].type != LEX_TOKEN_STR)
	{
		SystemError("[Err] LexGetChr \\0");
	}

	return Lex[id].str[0];
}
//---------------------------------------------------------------------------
IWRAM_CODE void LexGetToken(u32 id, bool isChr)
{
	char buf[LEX_MAX_STR_CNT];
	s32  num, cnt;

	ST_LEX* p = &Lex[id];
	p->pOld = p->pCur;

st: ;

	u32 type = LexGetType(*p->pCur);

	switch(type)
	{
	// 󔒁A^u
	case LEX_TOKEN_SPACE:
		p->pCur++;
		goto st;

	// R}
	case LEX_TOKEN_CUM:
		p->pCur++;
		goto st;

	// CR
	case LEX_TOKEN_CR:
		p->pCur += 2;
		goto st;

	// LF
	case LEX_TOKEN_LF:
		p->pCur++;
		goto st;

	// Rg
	case LEX_TOKEN_COMMENT:
		LexSkipLine(id);
		goto st;

	// I[
	case LEX_TOKEN_END:
		p->type = type;
		break;

	// 
	case LEX_TOKEN_NUM:
		num = 0;
		cnt = 0;

		do {
			num = (num * 10) + (*p->pCur++ - '0');
			cnt++;

		} while((LexGetType(*p->pCur) == LEX_TOKEN_NUM) && (cnt <= 8));

		p->type = type;
		p->num  = num;
		break;

	// _uNH[gt
	case LEX_TOKEN_DQUOTE:
		cnt = 0;
		p->pCur++;

		do {
			buf[cnt++] = *p->pCur++;

		} while((LexGetType(*p->pCur) != LEX_TOKEN_DQUOTE) && (cnt < LEX_MAX_STR_CNT-1) && (isChr == FALSE));

		buf[cnt] = '\0';
		p->pCur++;

		p->type = LEX_TOKEN_STR;
		p->num  = cnt;
		_Strncpy(p->str, buf, cnt+1);
		break;

	// 
	case LEX_TOKEN_STR:
		cnt = 0;

		do {
			buf[cnt++] = *p->pCur++;

		} while((LexGetType(*p->pCur) != LEX_TOKEN_SPACE) && (LexGetType(*p->pCur) != LEX_TOKEN_CR) && (LexGetType(*p->pCur) != LEX_TOKEN_LF) && (cnt < LEX_MAX_STR_CNT-1) && (isChr == FALSE));

		buf[cnt] = '\0';

		p->type = type;
		p->num  = cnt;
		_Strncpy(p->str, buf, cnt+1);
		break;
	}

	LexSkipLf(id);
}
//---------------------------------------------------------------------------
IWRAM_CODE u32 LexGetType(char chr)
{
	if(chr == ' ' || chr == '\t')
	{
		return LEX_TOKEN_SPACE;
	}
	else if(chr == CR)
	{
		return LEX_TOKEN_CR;
	}
	else if(chr == LF)
	{
		return LEX_TOKEN_LF;
	}
	else if(chr == ',')
	{
		return LEX_TOKEN_CUM;
	}
	else if(chr == '#')
	{
		return LEX_TOKEN_COMMENT;
	}
	else if(chr == '\"')
	{
		return LEX_TOKEN_DQUOTE;
	}
	else if(chr == '\0')
	{
		return LEX_TOKEN_END;
	}
	else if(_IsDigit(chr) == TRUE)
	{
		return LEX_TOKEN_NUM;
	}

	return LEX_TOKEN_STR;
}
//---------------------------------------------------------------------------
IWRAM_CODE void LexSkipLine(u32 id)
{
	ST_LEX* p = &Lex[id];

	while(*p->pCur != '\0')
	{
		if(*p->pCur == CR)
		{
			p->pCur += 2;
			return;
		}

		if(*p->pCur == LF)
		{
			p->pCur += 1;
			return;
		}

		p->pCur++;
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void LexSkipLf(u32 id)
{
	ST_LEX* p = &Lex[id];

	if(*p->pCur == LF)
	{
		p->pCur++;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void LexUnGet(u32 id)
{
	Lex[id].pOld = Lex[id].pCur;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool LexIsEnd(u32 id)
{
	return (Lex[id].pCur == '\0') ? TRUE : FALSE;
}
