#include "mode3.h"
#include "bios_arm.h"


//---------------------------------------------------------------------------
ST_MODE3 Mode3 EWRAM_BSS;


//---------------------------------------------------------------------------
EWRAM_CODE void Mode3Init(void)
{
	BiosExecFixClear(&Mode3, sizeof(ST_MODE3));

	REG_DISPCNT = MODE_3 | BG2_ON;
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3Exec(void)
{
	if(Mode3.isDraw == TRUE)
	{
		BiosExecFast(Mode3.bg, (u16*)VRAM, MODE3_MAX_BUF_SIZE);
		Mode3.isDraw = FALSE;
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3DrawBg(u16* pImg)
{
	LZ77UnCompVram(pImg, Mode3.bg);
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3DrawChr(u16* pImg)
{
	LZ77UnCompVram(pImg, Mode3.chr);
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3DrawCrop(u32 sx, u32 sy, u32 cx, u32 cy, u16* pImg)
{
	LZ77UnCompVram(pImg, Mode3.chr);


	u16* p = Mode3.bg + (sy * SCREEN_WIDTH) + sx;
	u32  x, y;

	for(y=0; y<cy; y++)
	{
		for(x=0; x<cx; x++)
		{
			p[y*SCREEN_WIDTH + x] = Mode3.chr[y*cx + x];
		}
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3DrawBlend(u32 sx, u32 sy, u32 cx, u32 cy, u16* pImg, u16* pMask)
{
	LZ77UnCompVram(pImg, Mode3.chr);
	LZ77UnCompVram(pMask, Mode3.mask);

	u32 x, y;
	u32 ex = sx + cx;
	u32 ey = sy + cy;

	u16* pC = Mode3.chr;
	u8*  pM = Mode3.mask;

	for(y=sy; y<ey; y++)
	{
		for(x=sx; x<ex; x++)
		{
			if(*pM == 0x1f)
			{
				Mode3.bg[(y * SCREEN_WIDTH) + x] = *pC;
			}
			else if(*pM == 0x00)
			{
				// EMPTY
			}
			else
			{
				u32 a = *pM;
				u32 s = *pC;
				u32 d = Mode3.bg[(y * SCREEN_WIDTH) + x];

				u32 sR = (s & 0x001f);
				u32 sG = (s & 0x03e0);
				u32 sB = (s & 0x7c00);

				u32 dR = (d & 0x001f);
				u32 dG = (d & 0x03e0);
				u32 dB = (d & 0x7c00);

				u32 r = ((sR - dR) * a) / 32 + dR;
				u32 g = ((sG - dG) * a) / 32 + dG;
				u32 b = ((sB - dB) * a) / 32 + dB;

				Mode3.bg[(y * SCREEN_WIDTH) + x] = (b & 0x7c00) | (g & 0x03e0) | (r & 0x001f);
			}

			pM++;
			pC++;
		}
	}
}
//---------------------------------------------------------------------------
// XN[p
IWRAM_CODE void Mode3DrawBlendWhite(u32 sx, u32 sy, u32 cx, u32 cy, u32 ex, u32 ey, u32 alpha)
{
	u32 x, y;

	for(y=0; y<cy; y++)
	{
		for(x=0; x<cx; x++)
		{
			u32 s = Mode3.bg[((sy + y) * SCREEN_WIDTH) + (sx + x)];

			if(s != 0x7fff)
			{
				u32 r = (s & 0x001f) + (alpha);
				u32 g = (s & 0x03e0) + (alpha <<  5);
				u32 b = (s & 0x7c00) + (alpha << 10);

				if(r > 0x001f) r = 0x001f;
				if(g > 0x03e0) g = 0x03e0;
				if(b > 0x7c00) b = 0x7c00;

				Mode3.bg[((ey + y) * SCREEN_WIDTH) + (ex + x)] = (b & 0x7c00) | (g & 0x03e0) | (r & 0x001f);
			}
		}
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3Copy(u32 sx, u32 sy, u32 ex, u32 ey)
{
	_ASSERT(sx < SCREEN_WIDTH);
	_ASSERT(sy < SCREEN_HEIGHT);

	u32 x, y;

	for(y=sy; y<=ey; y++)
	{
		for(x=sx; x<=ex; x++)
		{
			((u16*)VRAM)[(y * SCREEN_WIDTH) + x] = Mode3.bg[(y * SCREEN_WIDTH) + x];
		}
	}

	_ASSERT(ex < SCREEN_WIDTH);
	_ASSERT(ey < SCREEN_HEIGHT);
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3Copy2(u32 sx, u32 sy, u32 cx, u32 cy, u32 ex, u32 ey)
{
	u16* p1 = (u16*)VRAM + (ey * SCREEN_WIDTH) + ex;
	u16* p2 = Mode3.bg   + (sy * SCREEN_WIDTH) + sx;
	u32  x, y;

	for(y=0; y<cy; y++)
	{
		for(x=0; x<cx; x++)
		{
			p1[y*SCREEN_WIDTH + x] = p2[y*SCREEN_WIDTH + x];
		}
	}
}
//---------------------------------------------------------------------------
// I[vjOpXN[
IWRAM_CODE void Mode3ScrollOp(u32 cnt)
{
	if(cnt == 0)
	{
		BiosExecFast(Mode3.bg, (u16*)VRAM, MODE3_MAX_BUF_SIZE);
		return;
	}

	if(cnt == 160)
	{
		BiosExecFast(Mode3.chr, (u16*)VRAM, MODE3_MAX_BUF_SIZE);
		return;
	}

	// {BiosExecgłA@ƃG~[^iBIOSj
	// ɂׂ̋ȏuĂ܂
	Mode3ScrollOpSub(Mode3.chr + (240 * 160) - (240 * cnt), (u16*)VRAM, (240 * cnt) * 2);
	Mode3ScrollOpSub(Mode3.bg, (u16*)VRAM + (240 * cnt), ((240 * 160) - (240 * cnt)) * 2);
}
//---------------------------------------------------------------------------
IWRAM_CODE void Mode3ScrollOpSub(u16* pSrc, u16* pDst, u32 size)
{
	size /= 2;

	while(size != 0)
	{
		*pDst++ = *pSrc++;
		*pDst++ = *pSrc++;
		size -= 2;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void Mode3SetDraw(bool isDraw)
{
	Mode3.isDraw = isDraw;
}
