#include "nv.h"
#include "nv2.h"
#include "libmy/key.h"
#include "file.h"
#include "history.h"
#include "img.h"
#include "menu.h"
#include "bgm.h"
#include "anime.h"
#include "se.h"
#include "img.h"


//---------------------------------------------------------------------------
ST_NV Nv;


//---------------------------------------------------------------------------
EWRAM_CODE void NvInit(void)
{
	_Memset(&Nv, 0x00, sizeof(ST_NV));

	NvSetAct(NV_ACT_PARSE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExec(void)
{
	if(ImgIsEnd() == FALSE || TextIsEnd() == FALSE || AnimeIsEnd() == FALSE || HistoryIsEnd() == FALSE || MenuIsEnd() == FALSE)
	{
		return;
	}

	if(Nv.act == NV_ACT_END)
	{
		return;
	}


	switch(Nv.act)
	{
	case NV_ACT_PARSE:
		NvExecParse();
		break;

	case NV_ACT_KEY:
		NvExecKey();
		break;

	case NV_ACT_KEY_WAIT:
		NvExecKeyWait();
		break;

	case NV_ACT_SELECT:
		NvExecSelect();
		break;

	case NV_ACT_END_READY:
		NvExecEndReady();
		break;

	default:
		SystemError("[Err] NvExec Nv.act = %d\n", Nv.act);
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecKey(void)
{
	if(NvExecKeySub() == TRUE)
	{
		TextSetLf();
		NvSetAct(NV_ACT_PARSE);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool NvExecKeySub(void)
{
	u16 cnt = KeyGetCnt();
	u16 trg = KeyGetTrg();

	TextShow();
	TextShowCur();

	if(Nv.isSkip == TRUE)
	{
		if(cnt & KEY_B)
		{
			Nv.isSkip = FALSE;
		}
		else
		{
			TextHideCur();
			return TRUE;
		}
	}

	if((trg & KEY_A) || (trg & KEY_DOWN) || (cnt & KEY_R))
	{
		TextHideCur();
		return TRUE;
	}
	else if((trg & KEY_LEFT) && HistoryIsEmpty() == FALSE)
	{
		TextHideCur();
		HistoryInitAct(TRUE);

		return FALSE;
	}
	else if(trg & KEY_B)
	{
		TextHideCur();
		MenuSetSystem();

		return FALSE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecKeyWait(void)
{
	u16 trg = KeyGetTrg();

	if((trg & KEY_A) || (trg & KEY_DOWN) || Nv.isSkip == TRUE)
	{
		NvSetAct(NV_ACT_PARSE);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecSelect(void)
{
	s32 i;

	switch(Nv.actStep)
	{
	case 0:
		TextClear();
		TextShow();
		Nv.actStep++;
		break;

	case 1:
		Nv.sel = -1;
		Nv.actStep++;
		break;

	case 2:
		TextClearXY();
		
		for(i=0; i<Nv.selCnt; i++)
		{
			if(Nv.sel == i)
			{
				TextDrawStr((u16*)Nv.selStr[i], _Strlen((char*)Nv.selStr[i]) / 2, FALSE);
			}
			else
			{
				TextDrawStrGray((u16*)Nv.selStr[i], _Strlen((char*)Nv.selStr[i]) / 2, FALSE);
			}
			TextSetLf();
		}
		Nv.actStep++;
		break;

	case 3: ;
		u16 trg = KeyGetTrg();

		if((trg & KEY_UP) && Nv.sel > 0)
		{
			Nv.sel--;
			Nv.actStep--;
		}
		else if((trg & KEY_DOWN) && (Nv.sel + 1) < Nv.selCnt)
		{
			Nv.sel++;
			Nv.actStep--;
		}
		else if((trg & KEY_A) && Nv.sel != -1)
		{
			Nv.actStep++;
		}
		else if((trg & KEY_LEFT) && HistoryIsEmpty() == FALSE)
		{
			HistoryInitAct(TRUE);
			Nv.actStep--;
		}
		else if(trg & KEY_B)
		{
			MenuSetSystem();
			Nv.actStep--;
		}
		break;

	case 4:
		TextClearDat();
		TextClearXY();

		NvSetVarNum(Nv.selVar, Nv.sel+1);
		Nv.isSkip = FALSE;
		NvSetAct(NV_ACT_PARSE);
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecEndReady(void)
{
	BgmStop();
	SeStop();

	HistoryInit();
	TextClear();
	TextShow();

	ImgClear();
	ImgSetBg("SIRO");
	ImgSetEffect(IMG_EFFECT_0);


	NvSetAct(NV_ACT_END);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSetAct(u32 act)
{
	Nv.act = act;
	Nv.actStep = 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSetDat(u32 no)
{
	char* pDat = (char*)FileGetDat(no);

	Nv.pHead = pDat;
	Nv.pCur  = pDat;
	Nv.no    = no;

	TRACEOUT("[Load: %d]\n", no);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSetSkip(bool is)
{
	Nv.isSkip = is;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSetVarNum(u32 idx, u32 val)
{
	u32 map = NvGetVarNumMap(idx);

	TRACEOUT("[SetVarNum[%d] = %d]\n", map, val);
	Nv.varNum[map] = val;
}
//---------------------------------------------------------------------------
EWRAM_CODE u32 NvGetVarNum(u32 idx)
{
	u32 map = NvGetVarNumMap(idx);

	TRACEOUT("[GetVarNum[%d] val:%d]\n", map, Nv.varNum[map]);
	return Nv.varNum[map];
}
//---------------------------------------------------------------------------
EWRAM_CODE u32 NvGetVarNumMap(u32 idx)
{
	u32 ret;

	if(idx == 2)
	{
		ret = 254;
	}
	else if(idx == 3)
	{
		ret = 255;
	}
	else
	{
		ret = idx - 300;
	}

	_ASSERT(ret < NV_MAX_VAR_NUM_CNT);
	return ret;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSetVarStr(u32 idx, char* pStr)
{
	_ASSERT(idx < NV_MAX_VAR_STR_CNT);
	TRACEOUT("[SetVarStr[%d] = \"%s\"]\n", idx, pStr);

	_Strncpy(Nv.varStr[idx], pStr, NV_MAX_STR_SIZE);
}
//---------------------------------------------------------------------------
EWRAM_CODE char* NvGetVarStr(u32 idx)
{
	_ASSERT(idx < NV_MAX_VAR_STR_CNT);
	TRACEOUT("[GetVarStr[%d] = \"%s\"]\n", idx, Nv.varStr[idx]);

	return Nv.varStr[idx];
}
//---------------------------------------------------------------------------
EWRAM_CODE bool NvIsEnd(void)
{
	return (Nv.act == NV_ACT_END) ? TRUE : FALSE;
}
