#include "nv2.h"
#include "nv.h"
#include "libmy/lex.h"
#include "img.h"
#include "text.h"
#include "bgm.h"
#include "menu.h"
#include "history.h"
#include "se.h"
#include "anime.h"


//---------------------------------------------------------------------------
ST_NV_PAT_TABLE NvPatTable[NV_MAX_PAT_TABLE_CNT] = {
	{ "\\",       (void*)NvExecParseKey      },
	{ "$",        (void*)NvExecParseKey2     },
	{ "@",        (void*)NvExecParseShow     },
	{ "img1",     (void*)NvExecParseImg1     },
	{ "img2",     (void*)NvExecParseImg2     },
	{ "img3",     (void*)NvExecParseImg3     },
	{ "img4",     (void*)NvExecParseImg4     },
	{ "img5",     (void*)NvExecParseImg5     },
	{ "jump1",    (void*)NvExecParseJump1    },
	{ "jump2",    (void*)NvExecParseJump2    },
	{ "jump3",    (void*)NvExecParseJump3    },
	{ "calendar", (void*)NvExecParseCalendar },
	{ "call2",    (void*)NvExecParseCall2    },
	{ "ret",      (void*)NvExecParseRet      },
	{ "var",      (void*)NvExecParseVar      },
	{ "str",      (void*)NvExecParseStr      },
	{ "bgm",      (void*)NvExecParseBgm      },
	{ "bgm_fade", (void*)NvExecParseBgmFade  },
	{ "wav",      (void*)NvExecParseWav      },
	{ "wav_stop", (void*)NvExecParseWavStop  },
	{ "wait",     (void*)NvExecParseWait     },
	{ "select",   (void*)NvExecParseSelect   },
	{ "fade",     (void*)NvExecParseFade     },
	{ "msg_hide", (void*)NvExecParseMsgHide  },
	{ "shake",    (void*)NvExecParseShake    },
	{ "flash",    (void*)NvExecParseFlash    },
	{ "title",    (void*)NvExecParseTitle    },
};

u8 NvSjisTable[] =
"[Ă"
"Ȃɂʂ˂̂͂Ђӂւق܂݂ނ߂";


//---------------------------------------------------------------------------
extern ST_NV Nv;


//---------------------------------------------------------------------------
IWRAM_CODE void NvExecParse(void)
{
	if(Nv.wait != 0)
	{
		Nv.wait--;
		return;
	}


	Nv.isLoop = TRUE;

	do
	{
		TRACEOUT("\n");
		TRACEOUT("[%x %x SEEN%03d.dat]\n", Nv.pCur - Nv.pHead, Nv.pCur, Nv.no);

		NvExecParseSub();
		NvExecParseSubPatch();

	} while(Nv.isLoop == TRUE);
}
//---------------------------------------------------------------------------
IWRAM_CODE void NvExecParseSub(void)
{
	char* p = NvGetStr();

	if(_IsSJIS(*p) == TRUE || _IsKana(*p) == TRUE)
	{
		TRACEOUT("[Msg]\n");
		NvMsg(p);

		return;
	}


	u32 i;

	for(i=0; i<NV_MAX_PAT_TABLE_CNT; i++)
	{
		if(_Strncmp(p, NvPatTable[i].pStr, NV_MAX_STR_SIZE) == 0)
		{
			break;
		}
	}
	_ASSERT(i < NV_MAX_PAT_TABLE_CNT);


	TRACEOUT("[%s]\n", p);
	NvPatTable[i].pFunc();
}
//---------------------------------------------------------------------------
IWRAM_CODE void NvExecParseSubPatch()
{
	u32 adr = Nv.pCur - Nv.pHead;
	u32 nextAdr = 0;
	u32 nextAni = 0;

	switch(Nv.no)
	{
	// SAYU_01
	case 269:
		if(adr == 0x495a)
		{
			nextAni = 15;
			nextAdr = 0x4a17;
		}
		break;

	// S_AYU2, S_AYU1
	case 414:
		if(adr == 0x1c96)
		{
			nextAni = 17;
			nextAdr = 0x1d11;
		}
		else if(adr == 0x2d35)
		{
			nextAni = 16;
			nextAdr = 0x2e37;
		}
		break;

	// S_SIO2
	case 424:
		if(adr == 0x3d5a)
		{
			nextAni = 18;
			nextAdr = 0x3d72;
		}
		break;

	default:
		// EMPTY
		break;
	}

	if(nextAdr != 0)
	{
		TRACEOUT("[Patch %x %x]\n", nextAdr, nextAni);
		AnimeSetDat(nextAni);

		Nv.pCur = Nv.pHead + nextAdr;
		Nv.isLoop = FALSE;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseKey(void)
{
	if(Nv.msgCnt == 0)
	{
		NvSetAct(NV_ACT_KEY_WAIT);
	}
	else
	{
		NvText();
		NvSetAct(NV_ACT_KEY);
	}

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseKey2(void)
{
	if(Nv.msgCnt == 0)
	{
		return;
	}

	NvText();
	NvSetAct(NV_ACT_KEY);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseShow(void)
{
	if(Nv.msgCnt == 0)
	{
		return;
	}

	NvText();

	Nv.wait = 60 * 5;
	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseImg1(void)
{
	char* p = NvGetStr();
	u32   v = NvGetNum();

	TRACEOUT("[%s %d]\n", p, v);

	ImgClear();
	ImgSetBg(p);
	NvEffect(v);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseImg2(void)
{
	char* p = NvGetStr();

	TRACEOUT("[%s]\n", p);

	ImgClear();
	ImgSetBg(p);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseImg3(void)
{
	char* p = NvGetStr();
	u32   v = NvGetNum();

	TRACEOUT("[%s %d]\n", p, v);

	ImgClear();
	ImgSetBg(p);
	NvEffect(v);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseImg4(void)
{
	u32 cnt = NvGetNum();
	u32 idx = NvGetNum();

	switch(cnt)
	{
	case 0:
		TRACEOUT("[%d]\n", cnt);
		break;

	case 1: ;
		char* p1 = NvGetStr();
		TRACEOUT("[%d %s]\n", cnt, p1);

		ImgSetChr(p1);
		break;

	case 2: ;
		char* p2;
		TRACEOUT("[%d ", cnt);

		p2 = NvGetStr();
		TRACEOUT("%s ", p2);
//		ImgSetBg2(p2);			// TODO d̂ŌĂяoĂ܂

		p2 = NvGetStr();
		TRACEOUT("%s]\n", p2);
		ImgSetChr(p2);
		break;

	default:
		SystemError("[Err] NvExecParseImg4 %d\n", cnt);
		break;
	}

	NvEffect(idx);
	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseImg5(void)
{
	char* p = NvGetStr();

	if(p[0] == '?')
	{
		return;
	}

	ImgClear();
	ImgSetBg(p);
	NvEffect(IMG_EFFECT_0);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseJump1(void)
{
	bool is = NvGetIfCondition();
	u32  v  = NvGetNum();

	if(is == FALSE)
	{
		Nv.pCur = Nv.pHead + v;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseJump2(void)
{
	u32 v = NvGetNum();

	if(v == 504)
	{
		// GfBOZbg
		AnimeSetDat(3);

		TextHide();
		TextClear();

		NvSetAct(NV_ACT_END_READY);
		Nv.isLoop = FALSE;
		return;
	}

	if(v == 2)
	{
		NvSetAct(NV_ACT_END_READY);
		Nv.isLoop = FALSE;
		return;
	}


	NvSetDat(v);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseJump3(void)
{
	Nv.pCur = Nv.pHead + NvGetNum();
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseCalendar(void)
{
	u32 month = NvGetVarNum(2);
	u32 day = NvGetVarNum(3);

	if(month == 1)
	{
		ImgSetCalendar(day);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseCall2(void)
{
	u32 v = NvGetNum();

	if(v >= 501)
	{
		AnimeSetDat(v - 501);
		Nv.isLoop = FALSE;

		return;
	}

	NvCallPush();
	NvSetDat(v);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseRet(void)
{
	NvCallPop();
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseVar(void)
{
	char* p = NvGetStr();
	u32 idx;
	u32 val;

	if(NvIsStr(p, "=") == TRUE)
	{
		idx = NvGetVar();
		val = NvGetNum2();
		goto End;
	}

	if(NvIsStr(p, "+=") == TRUE)
	{
		idx = NvGetVar();
		val = NvGetVarNum(idx) + NvGetNum2();
		goto End;
	}

	if(NvIsStr(p, "-=") == TRUE)
	{
		idx = NvGetVar();
		val = NvGetVarNum(idx) - NvGetNum2();
		goto End;
	}

	SystemError("[Err] NvExecParseVar %s\n", p);
	return;

End:
	NvSetVarNum(idx, val);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseStr(void)
{
	u32 idx = NvGetNum();
	char* p = NvGetStr();

	NvSetVarStr(idx, p);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseBgm(void)
{
	u32 no = NvGetNum();
	bool is = (NvGetNum() == 1) ? TRUE : FALSE;

	BgmPlay(no, is);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseBgmFade(void)
{
	NvGetNum();

	BgmStop();
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseWav(void)
{
	char* p = NvGetStr();

	SePlay(p, FALSE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseWavStop(void)
{
	SeStop();
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseWait(void)
{
	Nv.wait = Div(NvGetNum(), 60);		// (ms / vblank)

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseSelect(void)
{
	Nv.selCnt = 0;
	Nv.selVar = NvGetNum();
	NvSetVarNum(Nv.selVar, 0);

	do {
		char* p = NvGetStr();
		NvMsg(p);

		NvSelPush();
		Nv.msgCnt = 0;

	} while(NvIsStrNext("select_end") == FALSE);

	NvSkip();
	NvSetAct(NV_ACT_SELECT);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseFade(void)
{
	u32 v = NvGetNum();
	ImgClear();

	switch(v)
	{
	case 1:
		ImgSetBg("KURO");
		break;

	case 2:
		ImgSetBg("SIRO");
		break;

	default:
		SystemError("[Err] NvExecParseFade %d\n", v);
		break;
	}

	NvEffect(IMG_EFFECT_0);
	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseMsgHide(void)
{
	TextHide();

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseShake(void)
{
	NvGetNum();

	// EMPTY
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseFlash(void)
{
	ImgClear();
	ImgSetBg("SIRO");
	NvEffect(IMG_EFFECT_200);

	Nv.isLoop = FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvExecParseTitle(void)
{
	NvGetStr();

	// EMPTY
}
//---------------------------------------------------------------------------
IWRAM_CODE void NvMsg(char* p)
{
	for(;;)
	{
		if(*p == '\0')
		{
			Nv.msg[Nv.msgCnt] = '\0';

			TRACEOUT("[%s]\n", Nv.msg);
			_ASSERT(Nv.msgCnt+1 < NV_MAX_MSG_SIZE);

			return;
		}

		if(_IsSJIS(*p) == TRUE)
		{
			Nv.msg[Nv.msgCnt++] = *p++;
			Nv.msg[Nv.msgCnt++] = *p++;

			continue;
		}

		if(_IsKana(*p) == TRUE)
		{
			u32 idx = (*p++ - 0xA6) * 2;

			Nv.msg[Nv.msgCnt++] = NvSjisTable[idx];
			Nv.msg[Nv.msgCnt++] = NvSjisTable[idx+1];

			continue;
		}

		SystemError("[Err] NvMsg %s\n", p);
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void NvText(void)
{
	TextSetDraw((u16*)Nv.msg, Nv.msgCnt / 2);
	TextShow();

	Nv.msgCnt = 0;
}
//---------------------------------------------------------------------------
IWRAM_CODE bool NvGetIfCondition(void)
{
	u8  t[32];
	s16 n[32];
	u32 s, d;

	u32 cnt = 0;
	u32 depth = 0;

	TRACEOUT("[IfCondition]\n");

	for(;;)
	{
		char* p = NvGetStr();
//		TRACEOUT("  [%s]\n", p);

		if(NvIsStr(p, "(") == TRUE)
		{
			t[cnt] = '(';
			n[cnt] = 0;
			cnt++;

			depth++;
			continue;
		}

		if(NvIsStr(p, ")") == TRUE)
		{
			t[cnt] = ')';
			n[cnt] = 0;
			cnt++;

			depth--;
			if(depth <= 0)
			{
				break;
			}
			continue;
		}

		if(NvIsStr(p, "and") == TRUE)
		{
			t[cnt] = '&';
			n[cnt] = 0;
			cnt++;

			continue;
		}

		if(NvIsStr(p, "or") == TRUE)
		{
			t[cnt] = '|';
			n[cnt] = 0;
			cnt++;

			continue;
		}

		if(NvIsStr(p, "==") == TRUE)
		{
			s = NvGetNum2();
			d = NvGetNum2();

			t[cnt] = 'n';
			n[cnt] = (s == d) ? 1 : 0;
			cnt++;
			continue;
		}

		if(NvIsStr(p, "<=") == TRUE)
		{
			s = NvGetNum2();
			d = NvGetNum2();

			t[cnt] = 'n';
			n[cnt] = (s <= d) ? 1 : 0;
			cnt++;
			continue;
		}

		if(NvIsStr(p, ">=") == TRUE)
		{
			s = NvGetNum2();
			d = NvGetNum2();

			t[cnt] = 'n';
			n[cnt] = (s >= d) ? 1 : 0;
			cnt++;
			continue;
		}

		if(NvIsStr(p, ">") == TRUE)
		{
			s = NvGetNum2();
			d = NvGetNum2();

			t[cnt] = 'n';
			n[cnt] = (s > d) ? 1 : 0;
			cnt++;
			continue;
		}

		SystemError("[Err] NvGetIfCondition\n");
	}

	u32 i = 0;
	u32 j;

	while(cnt > 1)
	{
		if(t[i] == '(' && t[i+1] == 'n' && t[i+2] == ')')
		{
			t[i] = t[i+1];
			n[i] = n[i+1];

			for(j=i+2; j<cnt-1; j++) { t[j] = t[j+1]; n[j] = n[j+1]; }
			for(j=i+1; j<cnt-2; j++) { t[j] = t[j+1]; n[j] = n[j+1]; }
			cnt -= 2;
			i = 0;

			continue;
		}

		if(t[i] == 'n' && (t[i+1] == '|' || t[i+1] == '&') && t[i+2] == 'n')
		{
			if(t[i+1] == '|')
			{
				n[i] = n[i] | n[i+2];
			}
			else
			{
				n[i] = n[i] & n[i+2];
			}

			for(j=i+2; j<cnt-1; j++) { t[j] = t[j+1]; n[j] = n[j+1]; }
			for(j=i+1; j<cnt-2; j++) { t[j] = t[j+1]; n[j] = n[j+1]; }
			cnt -= 2;
			i = 0;

			continue;
		}

		i++;
	}

	TRACEOUT("[%d]\n", n[0]);
	return (n[0] == 1) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE char* NvGetStr(void)
{
	LexSetCur(LEX_ID_NV, Nv.pCur);

	char* p = LexGetStr(LEX_ID_NV);
	Nv.pCur = LexGetCur(LEX_ID_NV);

	if(NvIsStr(p, "s0") == TRUE)
	{
		return NvGetVarStr(0);
	}

	if(NvIsStr(p, "s1") == TRUE)
	{
		return NvGetVarStr(1);
	}

	return p;
}
//---------------------------------------------------------------------------
EWRAM_CODE u32 NvGetNum(void)
{
	LexSetCur(LEX_ID_NV, Nv.pCur);

	u32 n = LexGetNum(LEX_ID_NV);
	Nv.pCur = LexGetCur(LEX_ID_NV);

	return n;
}
//---------------------------------------------------------------------------
EWRAM_CODE u32 NvGetNum2(void)
{
	char* p = NvGetStr();
	u32 n = NvGetNum();

	if(NvIsStr(p, "v") == TRUE)
	{
		return NvGetVarNum(n);
	}

	if(NvIsStr(p, "n") == TRUE)
	{
		return n;
	}

	SystemError("[Err] NvGetNum2 [%s][%d]\n", p, n);
	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE u32 NvGetVar(void)
{
	char* p = NvGetStr();
	u32 n = NvGetNum();

	if(NvIsStr(p, "v") == TRUE)
	{
		return n;
	}

	SystemError("[Err] NvGetVar [%s][%d]\n", p, n);
	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSelPush(void)
{
	_ASSERT(Nv.selCnt < NV_MAX_SEL_CNT);
	_ASSERT(Nv.msgCnt < NV_MAX_SEL_STR_SIZE);

	_Strncpy((char*)Nv.selStr[Nv.selCnt], (char*)Nv.msg, NV_MAX_SEL_STR_SIZE);
	Nv.selCnt++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvCallPush(void)
{
	_ASSERT(Nv.callCnt < NV_MAX_CALL_CNT);

	Nv.callAdr[Nv.callCnt] = Nv.pCur - Nv.pHead;
	Nv.callNo[Nv.callCnt] = Nv.no;
	Nv.callCnt++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvCallPop(void)
{
	_ASSERT(Nv.callCnt != 0);

	Nv.callCnt--;
	NvSetDat(Nv.callNo[Nv.callCnt]);
	Nv.pCur = Nv.pHead + Nv.callAdr[Nv.callCnt];
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvEffect(u32 idx)
{
	TextHide();
	TextClear();

	if(Nv.isSkip == TRUE)
	{
		idx = IMG_EFFECT_200;
	}

	TRACEOUT("[effect %d]\n", idx);
	ImgSetEffect(idx);
}
//---------------------------------------------------------------------------
EWRAM_CODE void NvSkip(void)
{
	while(*Nv.pCur++ != 0x0a)
	{
		// EMPTY
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool NvIsStr(char* pSrc, char* pDst)
{
	return (_Strncmp(pSrc, pDst, _Strlen(pDst)) == 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool NvIsStrNext(char* pStr)
{
	LexSetCur(LEX_ID_NV, Nv.pCur);
	char* p = LexGetStr(LEX_ID_NV);

	return NvIsStr(p, pStr);
}
