#include <stdio.h>
#include <string.h>
#include "gba.h"


//---------------------------------------------------------------------------
u32  OpenBuf(char* pName, u8** pBuf);

void DecodePdt(char* pName, u8* pSrc, u32 srcSize);
void SaveFile(char* pName, u8* pDst, u32 dstSize, u32 cx, u32 cy);
void SaveFileMask(char* pName, u8* pDst, u32 dstSize, u32 cx, u32 cy);

void WriteS16(int num, BYTE *b, int index);
void WriteS32(int num, BYTE *b, int index);
s32  ReadS32(u8* b, u32 pos);


//---------------------------------------------------------------------------
int main(int argc,char **argv)
{
	if(argc != 2)
	{
		printf("err argc\n");
		exit(1);
	}

	u8* pSrc;
	u32 srcSize;

	srcSize = OpenBuf(argv[1], &pSrc);
	DecodePdt(argv[1], pSrc, srcSize);

	free(pSrc);
	return 0;
}
//---------------------------------------------------------------------------
u32 OpenBuf(char* pName, u8** pBuf)
{
	FILE* fp = fopen(pName, "rb");

	if(fp == NULL)
	{
		printf("err filename\n");
		exit(1);
	}

	fpos_t fsize;
	fseek(fp,0,SEEK_END);
	fgetpos(fp, &fsize);
	fseek(fp,0,SEEK_SET);

	*pBuf = (u8*)calloc(fsize, sizeof(u8));
	fread(*pBuf, sizeof(u8), fsize, fp);

	fclose(fp);
	return fsize;
}
//---------------------------------------------------------------------------
void DecodePdt(char* pName, u8* pSrc, u32 srcSize)
{
	if(memcmp(pSrc, "PDT10", 5) != 0)
	{
		printf("err pdt10: %s\n", pName);
		exit(1);
	}

	printf("pdtbmp ... %s\n", pName);


	s32 fileSize = ReadS32(pSrc,  8);
	s32 imgCx    = ReadS32(pSrc, 12);
	s32 imgCy    = ReadS32(pSrc, 16);
	s32 maskPtr  = ReadS32(pSrc, 28);
	s32 imgSize  = imgCx * imgCy;

	u8* pMask = NULL;
	u8* pDat  = NULL;

	s32 i;

	u8* pCur;
	s32 curCnt;
	u8* pEnd;
	u8* pBuf;
	s32 bit;

	s32 cnt;
	s32 num;
	u8* pRep;
	u8  flag;

	// Mask$B$,$"$k>l9g!"JLL>$G%U%!%$%kJ]B8$7$^$9(B
	if(maskPtr != 0)
	{
		pMask = (u8*)calloc(imgSize, sizeof(u8));

		if(pMask == NULL)
		{
			printf("err pMask\n");
			exit(1);
		}

		curCnt = fileSize - maskPtr;
		pCur   = pSrc + maskPtr;
		pBuf   = pMask;
		pEnd   = pMask + imgSize;
		bit    = 0;

		while((pBuf < pEnd) && (curCnt > 0))
		{
			if(!bit)
			{
				bit = 8;
				flag = *pCur++;
				curCnt--;
			}

			if(flag & 0x80)
			{
				*pBuf++ = *pCur++;
				curCnt--;
			}
			else
			{
				cnt     = (*pCur++)+2;
				pRep    = pBuf-((*pCur++)+1);
				curCnt -= 2;

				for(i=0; (i < cnt) && (pBuf < pEnd); i++)
				{
					*pBuf++ = *pRep++;
				}
			}

			bit--;
			flag <<= 1;
		}

		SaveFileMask(pName, pMask, imgSize, imgCx, imgCy);
		free(pMask);
	}

	// PDT10$B7A<0(B


	// RGB$B3F(B1$B%P%$%H(B
	imgSize *= 3;
	pDat = (u8*)calloc(imgSize, sizeof(u8));

	if(pDat == NULL)
	{
		printf("err pDat\n");
		exit(1);
	}

	pCur   = pSrc + 32;
	curCnt = (maskPtr != 0) ? maskPtr-32 : fileSize-32;
	pBuf   = pDat;
	pEnd   = pDat + imgSize;
	bit    = 0;

	while((pBuf < pEnd) && (curCnt > 0))
	{
		if(!bit)
		{
			bit = 8;
			flag = *pCur++;
			curCnt--;
		}

		if(flag & 0x80)
		{
			*pBuf++ = *pCur++;
			*pBuf++ = *pCur++;
			*pBuf++ = *pCur++;
			curCnt -= 3;
		}
		else
		{
			num  = *pCur++;
			num += ((*pCur++)<<8);			// $B0z?t$O%j%H%k%(%s%G%#%"%s$N(BWORD
			curCnt -= 2;

			cnt  = ((num&15)+1)*3;			// $B2<0L(B4bit$B$,%3%T!<?t(B. 2$B2s$,:G>.C10L!)(B
			pRep = pBuf-((num>>4)+1)*3;		// buf$B$O<!$N=q$-9~$_0LCV$r;X$7$F$k$N$G!"(B1$B%P%$%HB?$a$KLa$k(B

			for(i=0; (i < cnt) && (pBuf < pEnd); i++)
			{
				*pBuf++ = *pRep++;
			}
		}

		bit--;
		flag <<= 1;
	}

	SaveFile(pName, pDat, imgSize, imgCx, imgCy);
	free(pDat);
}
//---------------------------------------------------------------------------
void SaveFile(char* pName, u8* pDst, u32 dstSize, u32 cx, u32 cy)
{
	char dstName[64];
	strcpy(dstName, pName);

	char* p = memchr(dstName, '.', 64);
	strcpy(p, ".bmp");

	FILE* fp = fopen(dstName, "wb");
	if(fp == NULL)
	{
		printf("err save file: %s\n", dstName);
		exit(1);
	}


	int i,x,y;
	BYTE bmp_fileh[14];
	BYTE bmp_infoh[40];

	memset(bmp_fileh, 0, sizeof(bmp_fileh));
	memset(bmp_infoh, 0, sizeof(bmp_infoh));

	bmp_fileh[0] = 'B';
	bmp_fileh[1] = 'M';
	WriteS32(sizeof(bmp_fileh) + sizeof(bmp_infoh) + cx * cy * 3, bmp_fileh, 2);
	WriteS32(sizeof(bmp_fileh) + sizeof(bmp_infoh), bmp_fileh, 10);
	WriteS32(sizeof(bmp_infoh), bmp_infoh, 0);
	WriteS32(cx, bmp_infoh, 4);
	WriteS32(cy, bmp_infoh, 8);
	WriteS16(1,  bmp_infoh, 12);
	WriteS16(24, bmp_infoh, 14);

	fwrite(&bmp_fileh, sizeof(bmp_fileh), 1, fp);
	fwrite(&bmp_infoh, sizeof(bmp_infoh), 1, fp);

	for(y=cy-1; y>=0; y--)
	{
		for(x=0; x<cx; x++)
		{
			fputc(pDst[y*cx*3 + x*3 + 0], fp);
			fputc(pDst[y*cx*3 + x*3 + 1], fp);
			fputc(pDst[y*cx*3 + x*3 + 2], fp);
		}

		if(((cx*3) % 4) != 0)
		{
			for(i=0;i<4-((cx*3) % 4);i++)
			{
				fputc(0, fp);
			}
		}
	}

	fclose(fp);
}
//---------------------------------------------------------------------------
void SaveFileMask(char* pName, u8* pDst, u32 dstSize, u32 cx, u32 cy)
{
	char dstName[64];
	strcpy(dstName, pName);

	char* p = memchr(dstName, '.', 64);
	strcpy(p, "_M.bmp");

	FILE* fp = fopen(dstName, "wb");
	if(fp == NULL)
	{
		printf("err save file: %s\n", dstName);
		exit(1);
	}


	int i, y;
	BYTE bmp_pal[256*4];
	BYTE *b;

	BYTE bmp_fileh[14];
	BYTE bmp_infoh[40];

	memset(bmp_fileh, 0, sizeof(bmp_fileh));
	memset(bmp_infoh, 0, sizeof(bmp_infoh));

	bmp_fileh[0] = 'B';
	bmp_fileh[1] = 'M';

	WriteS32(sizeof(bmp_fileh) + sizeof(bmp_infoh) + 256*4 + cx * cy, bmp_fileh, 2);
	WriteS32(sizeof(bmp_fileh) + sizeof(bmp_infoh) + 256*4, bmp_fileh, 10);
	WriteS32(sizeof(bmp_infoh), bmp_infoh, 0);
	WriteS32(cx, bmp_infoh, 4);
	WriteS32(cy, bmp_infoh, 8);
	WriteS16(1, bmp_infoh,  12);
	WriteS16(8, bmp_infoh,  14);

	for(i=0;i<256;i++)
	{
		bmp_pal[i*4+0] = i;
		bmp_pal[i*4+1] = i;
		bmp_pal[i*4+2] = i;
		bmp_pal[i*4+3] = 0;
	}

	fwrite(&bmp_fileh, sizeof(bmp_fileh), 1, fp);
	fwrite(&bmp_infoh, sizeof(bmp_infoh), 1, fp);
	fwrite(&bmp_pal, 256*4, 1, fp);

	for(y=cy-1; y>=0; y--)
	{
		b = &pDst[y*cx];
		fwrite(b, cx, 1, fp);

		if((cx % 4) != 0)
		{
			for(i=0;i<4-(cx % 4);i++)
			{
				fputc(0, fp);
			}
		}
	}

	fclose(fp);
}
//---------------------------------------------------------------------------
void WriteS16(int num, BYTE *b, int index)
{
	s32 c0, c1;

	num %= 65536;
	c0 = num % 256;
	c1 = num / 256;

	b[index] = c0;
	b[index+1] = c1;
}
//---------------------------------------------------------------------------
void WriteS32(int num, BYTE *b, int index)
{
	s32 c0, c1, c2, c3;

	c0 = num & 255;
	c1 = (num >> 8) & 255; 
	c2 = (num >> 16) & 255;
	c3 = (num >> 24) & 255;

	b[index] = c0; b[index+1] = c1, b[index+2] = c2, b[index+3] = c3;
}
//---------------------------------------------------------------------------
s32 ReadS32(u8* b, u32 pos)
{
	s32 ret = 0;

	ret += (b[pos]);
	ret += (b[pos+1] <<  8);
	ret += (b[pos+2] << 16);
	ret += (b[pos+3] << 24);

	return ret;
}
