#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "gba.h"


//---------------------------------------------------------------------------
#define MAX_ADR_BUF_NUM			10000000
#define MAX_JUMP_BUF_NUM		10000

u8*   databuf;
long  headpos;
long  curpos;
u8    cmd;
u8    subcmd;
bool  isEnd;
FILE* out;
char  savelog[1024];
bool isAdrWrite;


long srcAdr[MAX_ADR_BUF_NUM];
long dstAdr[MAX_ADR_BUF_NUM];
int  adrCnt;

long srcJumpAdr[MAX_JUMP_BUF_NUM];
long dstJumpAdr[MAX_JUMP_BUF_NUM];
int  jumpCnt;

//---------------------------------------------------------------------------

void SeenLoadBuffer(char* pName);
void SeenLoadBufferFree(void);

void SeenSaveFopen(char* pName);
void SeenSaveFclose(void);
void SeenSaveLog(char* format, ...);
void SeenSaveLogClear(void);
void SeenSaveLogWrite(void);

void SeenPushAdr(long curSrc, long curDst);
void SeenPushAdrJump(long curSrc, long curDst);
void SeenCalcAdr(void);

void Decode(void);
void DecodeSub(void);
void DecodeSkip(char* pSig);
bool DecodeIsSig(char* pS, char* pD);

void D00(void);
void D01(void);
void D02(void);
void D03(void);
void D04(void);
void D0b(void);
void D0c(void);
void D0e(void);
void D13(void);
void D15(void);
void D16(void);
void D17(void);
void D19(void);
void D1b(void);
void D1c(void);
void D20(void);
void D31(void);
void D37_57(void);
void D58(void);
void D59(void);
void D5b(void);
void D5c(void);
void D60(void);
void D63(void);
void D64_6a(void);
void D6d(void);
void D73(void);
void Dfe_ff(void);
void Error(void);

void  strcpy_avg(char* dst, char* src);
int   ReadFormattedText(char* b, int* attr);
void  ReadText(char* buf);
int   ReadValue(void);
int   GetInt(void);
int   ReadInt(int pos);
void  Skip(void);
void  DecodeConditions(void);
char* Flag_GetStr(int i);


//---------------------------------------------------------------------------
int main(int argc,char **argv)
{
	if(argc != 2)
	{
		printf("err argc\n");
		exit(1);
	}


	SeenLoadBuffer(argv[1]);
	SeenSaveFopen(argv[1]);

	printf("seendec ... %s\n", argv[1]);
	Decode();

	SeenCalcAdr();
	SeenSaveFclose();
	SeenLoadBufferFree();

	return 0;
}
//---------------------------------------------------------------------------
void SeenLoadBuffer(char* pName)
{
	FILE* fp = fopen(pName, "rb");

	if(fp == NULL)
	{
		printf("err openname\n");
		exit(1);
	}

	fpos_t fsize;
	fseek(fp,0,SEEK_END);
	fgetpos(fp, &fsize);
	fseek(fp,0,SEEK_SET);

	databuf = (unsigned char*)calloc(fsize, sizeof(unsigned char));
	fread(databuf, sizeof(unsigned char), fsize, fp);
	fclose(fp);
}
//---------------------------------------------------------------------------
void SeenLoadBufferFree(void)
{
	free(databuf);
}
//---------------------------------------------------------------------------
void SeenSaveFopen(char* pName)
{
	char buf[1024];
	strcpy(buf, pName);

	char* p = memchr(buf, '.', 1024);

	if(p == NULL)
	{
		printf("err savename: %s\n", pName);
		exit(1);
	}

	strcpy(p, ".dat");

	out = fopen(buf, "wb");
	if(out == NULL)
	{
		printf("err savefile: %s\n", buf);
		exit(1);
	}
}
//---------------------------------------------------------------------------
void SeenSaveFclose(void)
{
	_ASSERT(out != NULL);
	fclose(out);
}
//---------------------------------------------------------------------------
void SeenSaveLog(char* format, ...)
{
	char buf[1024];

	va_list ap;
	va_start(ap, format);
	vsnprintf(buf, 1024, format, ap);
	va_end(ap);

	strcat(savelog, buf);


	// patch
	if(DecodeIsSig(savelog, "img2 \"DAYS\"") == TRUE)
	{
		strcpy(savelog, "calendar\n");
	}
	else if(DecodeIsSig(savelog, "img4 2 60 \"CGYK00\" 1 \"CGAY22AN\"") == TRUE)
	{
		strcpy(savelog, "img4 2 60 \"CGYK00\" \"CGAY22AN\" \n");
	}
	else if(DecodeIsSig(savelog, "img4 1 26 4 \"FGMI18\"") == TRUE)
	{
		strcpy(savelog, "img1 \"FGMI18\" 26 \n");
	}
	else if(DecodeIsSig(savelog, "img1 \"?\"") == TRUE)
	{
		strcpy(savelog, "@\n");
	}
}
//---------------------------------------------------------------------------
void SeenSaveLogClear(void)
{
	if(isAdrWrite == TRUE)
	{
		jumpCnt--;
	}
	isAdrWrite = FALSE;

	savelog[0] = '\0';
}
//---------------------------------------------------------------------------
void SeenSaveLogWrite(void)
{
	fprintf(out, savelog);
	savelog[0] = '\0';

	if(isAdrWrite == TRUE)
	{
		dstJumpAdr[jumpCnt-1] = ftell(out) - sizeof("xxxxxxxx");
	}
	isAdrWrite = FALSE;
}
//---------------------------------------------------------------------------
void SeenPushAdr(long curSrc, long curDst)
{
	_ASSERT(adrCnt < MAX_ADR_BUF_NUM);

	srcAdr[adrCnt] = curSrc;
	dstAdr[adrCnt] = curDst;
	adrCnt++;
}
//---------------------------------------------------------------------------
void SeenPushAdrJump(long curSrc, long curDst)
{
	_ASSERT(jumpCnt < MAX_JUMP_BUF_NUM);

	srcJumpAdr[jumpCnt] = curSrc;
	dstJumpAdr[jumpCnt] = curDst;
	jumpCnt++;
}
//---------------------------------------------------------------------------
void SeenCalcAdr(void)
{
	int i, j;

	for(i=0; i<jumpCnt; i++)
	{
		fseek(out, dstJumpAdr[i], SEEK_SET);

		for(j=0; j<adrCnt; j++)
		{
			if(srcJumpAdr[i] == srcAdr[j])
			{
				fprintf(out, "%08ld", dstAdr[j]);
				break;
			}
		}
		_ASSERT(j < adrCnt);

	}
}
//---------------------------------------------------------------------------
void Decode(void)
{
	int i;

	headpos = 0x63 + (ReadInt(0x18) * 4);
	curpos  = headpos;
	isEnd   = FALSE;

	do {
		DecodeSub();

		// VOl``FbNāAꕔ̃XNvgXLbv܂
		if(DecodeIsSig(savelog, "jump1 ( != b 1 b 1998 )") == TRUE)
		{
			// J_[`揈
			SeenSaveLogClear();

			for(i=0; i<7; i++)	// TODO Oɏo͂ĂȂ"\0"JEgĂ̂Œ
			{
				DecodeSub();
				SeenSaveLogClear();
			}

			// img2
			DecodeSub();
			SeenSaveLogWrite();

			DecodeSkip("jump1 ( != v 2 n 0 and != b 1998 n 0 )");

			// calendar
			DecodeSub();
			SeenSaveLogWrite();

			for(i=0; i<39; i++)
			{
				DecodeSub();
				SeenSaveLogClear();
			}

			DecodeSkip("jump3 ");
		}
		if(DecodeIsSig(savelog, "jump1 ( == b 1994 n 1 )") == TRUE)
		{
			// ǃtȌWvidcpHj
			SeenSaveLogClear();
		}
		else if(DecodeIsSig(savelog, "jump1 ( == b 1995 n 1 )") == TRUE)
		{
			// ǃtȌWviSdcpj
			SeenSaveLogClear();
		}
		else if(DecodeIsSig(savelog, "jump1 ( == b") == TRUE)
		{
			// ǃtȌWv
			SeenSaveLogClear();

			DecodeSkip("");
			DecodeSkip("");
			DecodeSkip("jump");
			DecodeSkip("");
			DecodeSkip("jump");
		}
		else if(DecodeIsSig(savelog, "var = b") == TRUE)
		{
			// ǃtÕZbg
			SeenSaveLogClear();
		}
		else if(DecodeIsSig(savelog, "var = v 0 n 0") == TRUE)
		{
			// prs
			SeenSaveLogClear();
		}
		else if(DecodeIsSig(savelog, "var = v 10") == TRUE)
		{
			// prs
			SeenSaveLogClear();
		}
		else if(DecodeIsSig(savelog, "var = v 11") == TRUE)
		{
			// prs
			SeenSaveLogClear();
		}
		else
		{
			SeenSaveLogWrite();
		}

	} while(isEnd == FALSE);
}
//---------------------------------------------------------------------------
void DecodeSub(void)
{
//	printf("\n");
//	printf("cur:0x%x cmd:0x%x\n", (int)curpos, databuf[curpos]);

	SeenPushAdr(curpos, ftell(out));
	cmd = databuf[curpos++];

	switch(cmd)
	{
	case 0x00:
		D00();
		break;

	case 0x01:
		D01();
		break;

	case 0x02:
		D02();
		break;

	case 0x03:
		D03();
		break;

	case 0x04:
		D04();
		break;

	case 0x0b:
		D0b();
		break;

	case 0x0c:
		D0c();
		break;

	case 0x0e:
		D0e();
		break;

	case 0x13:
		D13();
		break;

	case 0x15:
		D15();
		break;

	case 0x16:
		D16();
		break;

	case 0x17:
		D17();
		break;

	case 0x19:
		D19();
		break;

	case 0x1b:
		D1b();
		break;

	case 0x1c:
		D1c();
		break;

	case 0x20:
		D20();
		break;

	case 0x31:
		D31();
		break;

	case 0x37:
	case 0x39:
	case 0x3B:
	case 0x3C:
	case 0x3D:
	case 0x3E:
	case 0x3F:
	case 0x40:
	case 0x41:
	case 0x42:
	case 0x43:
	case 0x49:
	case 0x4a:
	case 0x4b:
	case 0x4c:
	case 0x4d:
	case 0x4e:
	case 0x4f:
	case 0x50:
	case 0x51:
	case 0x56:
	case 0x57:
		D37_57();
		break;

	case 0x58:
		D58();
		break;

	case 0x59:
		D59();
		break;

	case 0x5b:
		D5b();
		break;

	case 0x5c:
		D5c();
		break;

	case 0x60:
		D60();
		break;

	case 0x63:
		D63();
		break;

	case 0x64:
	case 0x65:
	case 0x67:
	case 0x68:
	case 0x69:
	case 0x6a:
		D64_6a();
		break;

	case 0x6d:
		D6d();
		break;

	case 0x73:
		D73();
		break;

	case 0xfe:
	case 0xff:
		Dfe_ff();
		break;

	default:
		Error();
		break;
	}
}
//---------------------------------------------------------------------------
void DecodeSkip(char* pSig)
{
	DecodeSub();

	if(pSig[0] != '\0')
	{
//		printf("%s == %s\n", savelog, pSig);
		_ASSERT(strncmp(savelog, pSig, strlen(pSig)) == 0);
	}
	else
	{
		_ASSERT(strlen(savelog) == 0);
	}

	SeenSaveLogClear();
}
//---------------------------------------------------------------------------
bool DecodeIsSig(char* pS, char* pD)
{
	return (strncmp(pS, pD, strlen(pD)) == 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
// XNvgI
void D00(void)
{
	SeenSaveLog("end\n");
	isEnd = TRUE;
}
//---------------------------------------------------------------------------
// NbN҂BeLXgNA
void D01(void)
{
	// GBAł̓eLXgNAȂ\
	SeenSaveLog("\\\n");
}
//---------------------------------------------------------------------------
// s
void D02(void)
{
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x01:		// sACfg0ɖ߂
	case 0x02:		// sACfg͂̂܂
		SeenSaveLog("$\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
// NbN҂BNbN̓bZ[WNAɑď
void D03(void)
{
	SeenSaveLog("\\\n");
}
//---------------------------------------------------------------------------
void D04(void)
{
	subcmd = databuf[curpos];

	switch(subcmd)
	{
	// EBhEBbZ[Wobt@NA
	case 0x01:
		curpos++;
		SeenSaveLog("msg_hide\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D0e(void)
{
	int  data;
	char buf[1024];

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	// BGM[vĐ
	case 0x01:
		ReadText(buf);
		SeenSaveLog("bgm %s 1\n", buf);
		return;

	// BGMPĐ
	case 0x03:
		ReadText(buf);
		SeenSaveLog("bgm %s 0\n", buf);
		return;

	// BGM FadeOut / Stop
	case 0x10:
		data = ReadValue();
		SeenSaveLog("bgm_fade %d\n", data);
		return;

	// BGM Stop
	case 0x11:
		SeenSaveLog("bgm_stop\n");
		return;

	// WAVPĐ
	case 0x30:
		ReadText(buf);
		SeenSaveLog("wav \"%s\"\n", buf);
		return;

	// WAV~
	case 0x36:
		SeenSaveLog("wav_stop\n");
		return;

	// WAV~2
//	case 0x37:
//		data = ReadValue();		// prs
//		SeenSaveLog("wav_stop2 %d", data);
//		return;

	// WAV~3
	case 0x38:
		SeenSaveLog("wav_stop3\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D0b(void)
{
	typedef struct EFFECT {
		int sx1;
		int sy1;
		int sx2;
		int sy2;
		int dx;
		int dy;
		unsigned int cmd;
		unsigned int steptime;
		unsigned int prevtime;
		int srcpdt;
		int dstpdt;
		int mask;
		int step;
		int curcount;
		int arg1;
		int arg2;
		int arg3;
		int arg4;
		int arg5;
		int arg6;
	} EFFECT;

	char buf[1024], buf2[1024];
	int subcmd, idx, i;
	int srcx1, srcy1, srcx2, srcy2;
	int dstx1, dsty1, count, num;
	int arg1;
	EFFECT e;

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x01:
	case 0x03:
	case 0x05:
		ReadText(buf);
		idx = ReadValue();
		SeenSaveLog("img1 \"%s\" %d\n", buf, idx);
		return;

	case 0x02:
	case 0x04:
	case 0x06:
		ReadText(buf);
		e.sx1      = ReadValue();
		e.sy1      = ReadValue();
		e.sx2      = ReadValue();
		e.sy2      = ReadValue();
		e.dx       = ReadValue();
		e.dy       = ReadValue();
		e.steptime = ReadValue();
		e.cmd      = ReadValue();
		e.mask     = ReadValue();
		e.arg1     = ReadValue();
		e.arg2     = ReadValue();
		e.arg3     = ReadValue();
		e.step     = ReadValue();
		e.arg5     = ReadValue();
		e.arg6     = ReadValue();
		e.prevtime = 0;
		e.srcpdt   = 1;
		e.dstpdt   = 0;
		e.curcount = 0;

		SeenSaveLog("img5 \"%s\"\n", buf);
//		SeenSaveLog("clip %d %d %d %d  %d %d\n", e.sx1, e.sy1, e.sx2, e.sy2, e.dx, e.dy);
//		SeenSaveLog("effect2 %d %d %d %d %d %d %d %d\n", e.cmd, e.mask, e.arg1, e.arg2, e.arg3, e.step, e.arg5, e.arg6);
		return;

	case 0x10:
		ReadText(buf);
		idx = ReadValue();
		SeenSaveLog("img2 \"%s\"\n", buf);
		return;

	// PDT̏dˍ킹ix[X̓t@Cj
	case 0x22:
		count = databuf[curpos++];		// ɏd˂閇
		ReadText(buf);					// x[XƂȂt@Cit@Cj
		idx = ReadValue();				// \GtFNgԍifrom GAMEEXE.INIj
		SeenSaveLog("img3 ");
//		SeenSaveLog("img3 \"%s\" %d %d ", buf, count, idx);

		for(i=0; i<count; i++)
		{
			idx = databuf[curpos++];	// d˕̃^Cv
			ReadText(buf2);			// d˂t@Cit@Cj

			if(buf2[0])
			{
				strcpy(buf, buf2);
			}

			switch(idx)
			{
			case 0x01:
				SeenSaveLog("%d \"%s\" ", idx, buf);
				break;

			case 0x02:
				num = ReadValue();		// effect
				SeenSaveLog("\"%s\" %d ", buf, num);
				break;

			case 0x03:
				srcx1 = ReadValue();
				srcy1 = ReadValue();
				srcx2 = ReadValue();
				srcy2 = ReadValue();
				dstx1 = ReadValue();
				dsty1 = ReadValue();
//				SeenSaveLog("%d \"%s\" %d %d %d %d  %d %d ", idx, buf, srcx1, srcy1, srcx2, srcy2, dstx1, dsty1);
				SeenSaveLog("%d \"%s\" ", idx, buf);
				break;

			case 0x04:
				srcx1 = ReadValue();
				srcy1 = ReadValue();
				srcx2 = ReadValue();
				srcy2 = ReadValue();
				dstx1 = ReadValue();
				dsty1 = ReadValue();
				arg1  = ReadValue();
//				SeenSaveLog("%d \"%s\" %d %d %d %d  %d %d  %d ", idx, buf, srcx1, srcy1, srcx2, srcy2, dstx1, dsty1, arg1);
				SeenSaveLog("%d \"%s\" %d ", idx, buf, arg1);
				break;

			default:
				Error();
				break;
			}
		}
		SeenSaveLog("\n");
		return;

	// PDT̏dˍ킹ix[X͊PDTj
	case 0x24:
		count = databuf[curpos++];		// ɏd˂閇
		i = ReadValue();				// x[XƂȂt@CiPDT No.j
		idx = ReadValue();				// \GtFNgԍifrom GAMEEXE.INIj
//		SeenSaveLog("img4 %d %d %d ", i, count, idx);
		SeenSaveLog("img4 %d %d ", count, idx);

		for(i=0; i<count; i++)
		{
			idx = databuf[curpos++];	// d˕̃^Cv
			ReadText(buf2);				// d˂t@Cit@Cj

			if(buf2[0])
			{
				strcpy(buf, buf2);
			}

			switch(idx)
			{
			case 0x01:
				SeenSaveLog("\"%s\" ", buf);
				break;

			case 0x02:
				num = ReadValue();
				SeenSaveLog("\"%s\" ", buf);
				break;

			case 0x03:
				srcx1 = ReadValue();
				srcy1 = ReadValue();
				srcx2 = ReadValue();
				srcy2 = ReadValue();
				dstx1 = ReadValue();
				dsty1 = ReadValue();
//				SeenSaveLog("%d \"%s\" %d %d %d %d  %d %d ", idx, buf, srcx1, srcy1, srcx2, srcy2, dstx1, dsty1);
				SeenSaveLog("%d \"%s\" ", idx, buf);
				break;

			case 0x04:
				srcx1 = ReadValue();
				srcy1 = ReadValue();
				srcx2 = ReadValue();
				srcy2 = ReadValue();
				dstx1 = ReadValue();
				dsty1 = ReadValue();
				arg1  = ReadValue();
//				SeenSaveLog("%d \"%s\" %d %d %d %d  %d %d  %d ", idx, buf, srcx1, srcy1, srcx2, srcy2, dstx1, dsty1, arg1);
				SeenSaveLog("%d \"%s\" ", idx, buf);
				break;

			default:
				Error();
				break;
			}
		}
		SeenSaveLog("\n");
		return;

	case 0x30:
//		SeenSaveLog("redraw_clear\n");
		return;

	case 0x31:
		idx = ReadValue();
//		SeenSaveLog("redraw_delete %d\n", idx);
		return;

	case 0x33:
		idx = ReadValue();
//		SeenSaveLog("redraw_num_get v[%d]\n", idx);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D0c(void)
{
	char buf[1024];
	int idx;

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x10:
		ReadText(buf);
		idx = ReadValue();
		SeenSaveLog("anime_play \"%s\" %d\n", buf, idx);
		return;

	case 0x21:
//		SeenSaveLog("anime_stop\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D13(void)
{
	int idx;
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x01:
		idx = ReadValue();	// idx color
		SeenSaveLog("fade %d\n", idx);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D15(void)
{
	SeenSaveLog("jump1 ");
	DecodeConditions();

	int ptr = GetInt();
	SeenSaveLog("xxxxxxxx\n");

	SeenPushAdrJump(ptr + headpos, 0);
	isAdrWrite = TRUE;
}
//---------------------------------------------------------------------------
void D16(void)
{
	subcmd = databuf[curpos++];
	int ptr = ReadValue();

	switch(subcmd)
	{
	case 0x01:
		SeenSaveLog("jump2 %d\n", ptr);
		return;

	case 0x02:
		SeenSaveLog("call2 %d\n", ptr);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D17(void)
{
	subcmd = databuf[curpos++];
	int idx;

	switch(subcmd)
	{
	case 0x01:
		idx = ReadValue();
		SeenSaveLog("shake %d\n", idx);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D19(void)
{
	unsigned int waittime, waitcancelindex;

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x01:
		waittime = (unsigned int)ReadValue();		// ms
		SeenSaveLog("wait %d\n", waittime);
		return;

	case 0x02:
		waittime = (unsigned int)ReadValue();
		waitcancelindex = ReadValue();
		SeenSaveLog("wait2 %d %d\n", waittime, waitcancelindex);
		return;

	case 0x03:
//		SeenSaveLog("wait_reset\n");
		return;

	case 0x04:
		waittime = (unsigned int)ReadValue();		// ms
		SeenSaveLog("wait3 %d\n", waittime);
		return;

	case 0x05:
		waittime = (unsigned int)ReadValue();
		waitcancelindex = ReadValue();
		SeenSaveLog("wait4 %d %d\n", waittime, waitcancelindex);
		return;

	case 0x10:
//		SeenSaveLog("### Unknown Wait Cmd ($%02X-$%02X)\n", cmd, subcmd);
		return;

	case 0x11:
//		SeenSaveLog("### Unknown Wait Cmd ($%02X-$%02X)\n", cmd, subcmd);
		return;

	case 0x13:
//		SeenSaveLog("### Unknown Wait Cmd ($%02X-$%02X)\n", cmd, subcmd);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D1b(void)
{
	int ptr = GetInt();
	SeenSaveLog("call ");
	SeenSaveLog("xxxxxxxx\n");

	SeenPushAdrJump(ptr + headpos, 0);
	isAdrWrite = TRUE;
}
//---------------------------------------------------------------------------
void D1c(void)
{
	int ptr = GetInt();
	SeenSaveLog("jump3 ");
	SeenSaveLog("xxxxxxxx\n");

	SeenPushAdrJump(ptr + headpos, 0);
	isAdrWrite = TRUE;
}
//---------------------------------------------------------------------------
void D20(void)
{
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	// V[^[
	case 0x01:
		SeenSaveLog("return_local\n");
		return;

	// V[̃^[
	case 0x02:
		SeenSaveLog("ret\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D31(void)
{
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x02:
//		SeenSaveLog("### Unknown Cmd ($%02X-$%02X)\n", cmd, subcmd);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D37_57(void)
{
	int idx, data;

	idx  = ReadValue();
	data = ReadValue();

	SeenSaveLog("var ");

	switch(cmd)
	{
	case 0x37:
		SeenSaveLog("= b %d n %d\n", idx, data);
		return;

	case 0x39:
		SeenSaveLog("= b %d b %d\n", idx, data);
		return;

	case 0x3b:
		SeenSaveLog("= v %d n %d\n", idx, data);
		return;

	case 0x3c:
		SeenSaveLog("+= v %d n %d\n", idx, data);
		return;

	case 0x3d:
		SeenSaveLog("-= v %d n %d\n", idx, data);
		return;

	case 0x3e:
		SeenSaveLog("*= v %d n %d\n", idx, data);
		return;

	case 0x3f:
		SeenSaveLog("/= v %d n %d\n", idx, data);
		return;

	case 0x40:
		SeenSaveLog("%%= v %d n %d\n", idx, data);
		return;

	case 0x41:
		SeenSaveLog("&= v %d n %d\n", idx, data);
		return;

	case 0x42:
		SeenSaveLog("|= v %d n %d\n", idx, data);
		return;

	case 0x43:
		SeenSaveLog("^= v %d n %d\n", idx, data);
		return;

	case 0x49:
		SeenSaveLog("= v %d v %d\n", idx, data);
		return;

	case 0x4a:
		SeenSaveLog("+= v %d v %d\n", idx, data);
		return;

	case 0x4b:
		SeenSaveLog("-= v %d v %d\n", idx, data);
		return;

	case 0x4c:
		SeenSaveLog("*= v %d v %d\n", idx, data);
		return;

	case 0x4d:
		SeenSaveLog("/= v %d v %d\n", idx, data);
		return;

	case 0x4e:
		SeenSaveLog("%%= v %d v %d\n", idx, data);
		return;

	case 0x4f:
		SeenSaveLog("&= v %d v %d\n", idx, data);
		return;

	case 0x50:
		SeenSaveLog("|= v %d v %d\n", idx, data);
		return;

	case 0x51:
		SeenSaveLog("^= v %d v %d\n", idx, data);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D58(void)
{
	int selectflag, selectindex;
	int i, flag, attr;
	char buf[1024];

	selectflag = databuf[curpos++];
	i = 0;

	switch(selectflag)
	{
	case 0x01:		// ʑőI
	case 0x02:
		selectindex = ReadValue();		// ߂l̕ϐw
		SeenSaveLog("select %d\n", selectindex);

		if(databuf[curpos++] == 0x22)
		{
			if(!databuf[curpos])
			{
				curpos++;
			}

			do {
				flag = ReadFormattedText(buf, &attr);

				if(flag)
				{
					switch(attr & 0xff)
					{
					case 0x00:		// ́[܂
						SeenSaveLog("%s\n", buf);
						break;

					case 0x20:		// FtAI
					case 0x21:		// IƂďoȂH
					case 0x22:		// FtAI͕s
						Error();
						break;
					}
				}
				else
				{
					Error();
				}

			} while(databuf[curpos] != 0x23);

			curpos++;
		}

		SeenSaveLog("select_end\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D59(void)
{
	char buf[1024];
	int idx1, idx2;

	subcmd = databuf[curpos++];
	idx1 = ReadValue();

	switch(subcmd)
	{
	case 0x01:
		ReadText(buf);
		SeenSaveLog("str %d \"%s\"\n", idx1, buf);
		return;

	case 0x02:
		idx2 = ReadValue();
		SeenSaveLog("v[%d] = s[%d]\n", idx1, idx2);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D5b(void)
{
	int idx, data;

	subcmd = databuf[curpos++];
	idx = ReadValue();

	switch(subcmd)
	{
	case 1:
//		SeenSaveLog("Multiple Value Set");

		while(databuf[curpos])
		{
			data = ReadValue();
			SeenSaveLog("v[%d] = %d\n", idx, data);
			idx++;
		}
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D5c(void)
{
	int idx, idx2, data;

	subcmd = databuf[curpos++];
	idx  = ReadValue();
	idx2 = ReadValue();
	data = ReadValue();

	if(idx2 >= 2000)
	{
		idx2 = 1999;
	}

	switch(subcmd)
	{
	case 1:
		// idx-idx2 = data  gp
		SeenSaveLog("set_val %d %d %d\n", idx, idx2, data);
		return;

	case 2:
		// idx-idx2 = data  gp
		SeenSaveLog("set_bit %d %d %d\n", idx, idx2, data);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
// ̑
void D60(void)
{
	int idx, attr, data;
	char buf[1024];

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x02:
		idx = ReadValue();
		return;

	case 0x03:
		idx = ReadValue();
		return;

	case 0x04:
		idx = ReadFormattedText(buf, &attr);
		SeenSaveLog("title \"%s\"\n", buf);
		return;

	case 0x20:
		SeenSaveLog("Terminate\n");
		return;

	case 0x30:
		data = ReadValue();
		idx  = ReadValue();
		return;

	case 0x31:
		data = ReadValue();
		idx  = ReadValue();
		return;

	// Z[u^CgiȂj擾
	case 0x35:
		data = ReadValue();
		idx  = ReadValue();
		return;

	// Z[u擾
	case 0x36:
		data = ReadValue();
		idx  = ReadValue();
		return;

	// Z[u擾
	case 0x37:
		data = ReadValue();
		idx  = ReadValue();
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
// Graphics:Get/Putn
void D63(void)
{
	int data, n;

	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	case 0x20:
		n = databuf[curpos++];
		data = ReadValue();
//		SeenSaveLog("SnrPDT unknown %d %d\n", n, data);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D64_6a(void)
{
	int srcx1, srcy1, srcx2, srcy2;
	int dstx1, dsty1;
	int srcpdt, dstpdt;
	int r, g, b, flag, count;

	switch(cmd)
	{
	// ̈ɑ΂鑀֘AiNAƂF}XNƂj
	case 0x64:
		subcmd = databuf[curpos++];

		switch (subcmd)
		{
		case 0x15:
			srcx1 = ReadValue();
			srcy1 = ReadValue();
			srcx2 = ReadValue();
			srcy2 = ReadValue();
			srcpdt = ReadValue();
			r = ReadValue();
			g = ReadValue();
			b = ReadValue();
			count = ReadValue();

//			SeenSaveLog("col_Mask %d  %d %d  %d %d  %d %d %d  %d\n", srcpdt, srcx1, srcy1, srcx2, srcy2, r, g, b, count);
			break;

		default:
			Error();
			break;
		}
		return;

	// Unimpremented
	case 0x65:
		return;

	// obt@Rs[ƕ\֘A
	case 0x67:
		subcmd = databuf[curpos++];

		switch(subcmd)
		{
		case 0x00:
			Error();
//			srcx1 = ReadValue();
//			srcy1 = ReadValue();
//			srcx2 = ReadValue();
//			srcy2 = ReadValue();
//			srcpdt = ReadValue();
//			flag = ReadValue();
//			Sys_SnrPDT_CopyBackBuffer(srcx1, srcy1, srcx2, srcy2, srcpdt, 1/*flag*/);
//			SeenSaveLog("SnrPDT_CopyBackBuffer\n");
			break;

		// }XNRs[͗lH (Ribbon2ی̑I̐)
		// iKanon CG[hCGȂ݂ƁA}XN̓Rs[ȂۂEEEj
		// ڂ̕łRibbon2̂̕ŁAD
		case 0x01:
			srcx1 = ReadValue();
			srcy1 = ReadValue();
			srcx2 = ReadValue();
			srcy2 = ReadValue();
			srcpdt = ReadValue();
			dstx1 = ReadValue();
			dsty1 = ReadValue();
			dstpdt = ReadValue();
			flag = 0;
//			if ( Sys_Version()>=1704 ) flag = ReadValue();		// AVG32 New Version (>17D) Only
//			Sys_SnrPDT_Copy(srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, flag);
			SeenSaveLog("pdt_copy %d %d %d %d %d %d %d %d %d\n", srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, flag);
			break;

		case 0x02:
			srcx1 = ReadValue();
			srcy1 = ReadValue();
			srcx2 = ReadValue();
			srcy2 = ReadValue();
			srcpdt = ReadValue();
			dstx1 = ReadValue();
			dsty1 = ReadValue();
			dstpdt = ReadValue();
			flag = ReadValue();
//			if ( Sys_Version()>=1613 ) flag = ReadValue();			// AVG32 New Version (>16M) Only??
//			Sys_SnrPDT_MaskCopy(srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, flag);
//			SeenSaveLog("pdt_copy_mask %d %d %d %d %d %d %d %d %d\n", srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, flag);
			break;

		case 0x03:
			Error();
//			srcx1 = ReadValue();
//			srcy1 = ReadValue();
//			srcx2 = ReadValue();
//			srcy2 = ReadValue();
//			srcpdt = ReadValue();
//			dstx1 = ReadValue();
//			dsty1 = ReadValue();
//			dstpdt = ReadValue();
//			r = ReadValue();
//			g = ReadValue();
//			b = ReadValue();
//			Sys_SnrPDT_TransCopy(srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, r, g, b);
//			SeenSaveLog("SnrPDT_TransCopy\n");
			break;

		case 0x05:
			Error();
//			srcx1 = ReadValue();
//			srcy1 = ReadValue();
//			srcx2 = ReadValue();
//			srcy2 = ReadValue();
//			srcpdt = ReadValue();
//			dstx1 = ReadValue();
//			dsty1 = ReadValue();
//			dstpdt = ReadValue();
//			Sys_SnrPDT_Swap(srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt);
//			SeenSaveLog("SnrPDT_Swap\n");
			break;

		case 0x08:
			Error();
//			srcx1 = ReadValue();
//			srcy1 = ReadValue();
//			srcx2 = ReadValue();
//			srcy2 = ReadValue();
//			srcpdt = ReadValue();
//			dstx1 = ReadValue();
//			dsty1 = ReadValue();
//			dstpdt = ReadValue();
//			flag = ReadValue();
//			Sys_SnrPDT_DstMaskCopy(srcx1, srcy1, srcx2, srcy2, srcpdt, dstx1, dsty1, dstpdt, flag);
//			SeenSaveLog("SnrPDT_DstMaskCopy\n");
			break;

		case 0x11:		// SʃRs[B}XNobt@̓eRs[݂B
			srcpdt = ReadValue();
			dstpdt = ReadValue();
			flag = 0;
//			if ( Sys_Version()>=1704 ) flag = ReadValue();		// AVG32 New Version (>17D) Only
//			Sys_SnrPDT_AllCopy(srcpdt, dstpdt, flag);
//			SeenSaveLog("pdt_all_copy %d %d %d\n", srcpdt, dstpdt, flag);
			break;

		case 0x12:		// Sʃ}XNRs[
			Error();
//			srcpdt = ReadValue();
//			dstpdt = ReadValue();
//			flag = ReadValue();;
//			if ( Sys_Version()>=1613 ) flag = ReadValue();		// AVG32 New Version (>16M) Only
//			Sys_SnrPDT_MaskCopy(0, 0, 639, 479, srcpdt, 0, 0, dstpdt, flag);
//			SeenSaveLog("SnrPDT_MaskCopyAll\n");
			break;

		default:
			Error();
			break;
		}
		return;

	case 0x68:
		subcmd = databuf[curpos++];

		switch(subcmd)
		{
		// Fł߂H
		case 0x01:
			dstpdt = ReadValue();
			r = ReadValue();
			g = ReadValue();
			b = ReadValue();

			//  kanon̏ꍇAr g b255, 255, 255̂ݎgpBɂKvȂ
			SeenSaveLog("flash\n");
			break;

		default:
			Error();
			break;
		}
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D6d(void)
{
	int idx, idx2, idx3;
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
//	case 0x01:
//		Error();
//		return;

	case 0x02:
		idx = ReadValue();
		idx2 = ReadValue();
		idx3 = ReadValue();
//		SeenSaveLog("Mouse_SetState %d %d %d\n", idx, idx2, idx3);
		return;

	case 0x03:
//		SeenSaveLog("Mouse_ClickFlush\n");
		return;

	case 0x20:
//		SeenSaveLog("Mouse_Hide\n");
		return;

	case 0x21:
//		SeenSaveLog("Mouse_Show\n");
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void D73(void)
{
	int data;
	subcmd = databuf[curpos++];

	switch(subcmd)
	{
	// CtrlL[ɂXLbv̗L^
	case 0x1d:
		data = ReadValue();
//		SeenSaveLog("#CTRL_KEY = %d", data);
		return;

	// Set #RETURN_KEY_WAIT
	case 0x2d:		
		data = ReadValue();
//		SeenSaveLog("#RETURN_KEY_WAIT = %d", data);
		return;
	}

	Error();
}
//---------------------------------------------------------------------------
void Dfe_ff(void)
{
	char buf[1024];

	ReadText(buf);
	SeenSaveLog("%s\n", buf);
}
//---------------------------------------------------------------------------
void Error(void)
{
	printf("Error\n");
	exit(1);
}
//---------------------------------------------------------------------------
// ViǏ݈ʒu NullEnd  String ǂ
void strcpy_avg(char* dst, char* src)
{
	int i, j, k, l;
	unsigned char *d = (unsigned char*)dst, *s = (unsigned char*)src;

	for (i=0, j=0; (s[i])/*&&(j<255)*/; )
	{
		if ( (s[i]<0x80)||((s[i]>=0xa0)&&(s[i]<0xe0)) )
		{
			d[j++] = s[i++];
		}
		else if ( (s[i]==0x81)&&(s[i+1]==0x96) )
		{
			if( (s[i+2]==0x82)&&(s[i+3]>=0x61) )
			{
				d[j++] = 0x91;		// 
				d[j++] = 0x8a;
				d[j++] = 0x91;
				d[j++] = 0xf2;
				i += 4;
			} else if ( (s[i+2]==0x82)&&(s[i+3]>=0x60) ) {
				d[j++] = 0x97;		// S
				d[j++] = 0x53;
				d[j++] = 0x88;
				d[j++] = 0xea;
				i += 4;
			} else {
				d[j++] = s[i++];
				d[j++] = s[i++];
			}
		}
		else if(i == 0 && s[i] == 0x81 && s[i+1] == 0x79 && s[i+2] == 0x81 && s[i+3] == 0x7a)
		{
			// ǉ@yz̍폜Ή
			i += 4;
		}
		else
		{
			d[j++] = s[i++];
			d[j++] = s[i++];
		}
	}
	d[j] = 0;


	unsigned char tblA[] = "ĂƂȂɂʂ˂̂͂Ђӂւق܂݂ނ߂";
	unsigned char tblB[] = "ܦ";
	l = 0;

	for(i=0; i<j; i+=2)
	{
		for(k=0; k<sizeof(tblA); k+=2)
		{
			if(d[i] == tblA[k] && d[i+1] == tblA[k+1])
			{
				break;
			}
		}

		if(k >= sizeof(tblA))
		{
			d[l++] = d[i+0];
			d[l++] = d[i+1];
		}
		else
		{
			d[l++] = tblB[k/2];
		}
	}

	d[l] = 0;
}
//---------------------------------------------------------------------------
void ReadText(char* buf)
{
	int n;
	if (databuf[curpos] == '@') { // ϐ
		curpos++;
		n = ReadValue();
		strcpy_avg(buf, Flag_GetStr(n));
	} else {
		n = strlen((char*)&databuf[curpos]);
		strcpy_avg(buf, (char*)&databuf[curpos]);
		curpos += n+1;
	}
}
//---------------------------------------------------------------------------
int ReadFormattedText(char* b, int* attr)
{
	int ret = 1;			// TRUE
	unsigned char c, cc;
	int len, idx;
	char* buf = b;

	*attr = 0;
	do {
		c = databuf[curpos++];
		switch (c) {					// eLXgnR}ĥ܂ܒʂ̂ȁH
			case 0xfe:
			case 0xff:
				strcpy_avg(buf, (char*)&databuf[curpos]);
				len = strlen(buf);
				buf += len;
				len = strlen((char*)&databuf[curpos]);
				curpos += len+1;
				break;
			case 0xfd:
				idx = ReadValue();
				strcpy_avg(buf, Flag_GetStr(idx));
				buf += strlen(buf);
				break;
			case 0x28:
				Error();
				curpos--;
				DecodeConditions();
				if ( (ret!=0)&&(ret!=1) ) {				// 0/1ȊO̐lAgr[g
					*attr = ret;
					ret = 1;
				}
				break;
			case 0x10:
				cc = databuf[curpos++];
				switch (cc) {
					case 0x03:
						idx = ReadValue();
						strcpy_avg(buf, Flag_GetStr(idx));
						buf += strlen(buf);
						break;
				}
				break;
		}
	} while (c);
	*buf = 0;

	return ret;
}
//---------------------------------------------------------------------------
int ReadValue(void)
{
	int num, ret, i, l;

	num = (int)databuf[curpos];
	l = (num >> 4) & 7;
	ret = 0;

	for(i=curpos+l-1; i>curpos; i--)
	{
		ret <<= 8; 
		ret |= (int)databuf[i];
	}
	ret <<= 4;
	ret |= num&15;
	curpos += l;

	if(num & 0x80)
	{
//		Error();
		SeenSaveLog(" vg[%d] ", ret);

		return 0;
	}

	return ret;
}
//---------------------------------------------------------------------------
int GetInt(void)
{
	int ret;
	ret  =  (int)databuf[curpos++];
	ret |= ((int)databuf[curpos++])<<8;
	ret |= ((int)databuf[curpos++])<<16;
	ret |= ((int)databuf[curpos++])<<24;
	return ret;
}
//---------------------------------------------------------------------------
int ReadInt(int pos)
{
	int ret;
	ret  =  (int)databuf[pos++];
	ret |= ((int)databuf[pos++])<<8;
	ret |= ((int)databuf[pos++])<<16;
	ret |= ((int)databuf[pos++])<<24;
	return ret;
}
//---------------------------------------------------------------------------
void Skip(void)
{
	while(databuf[curpos++] != 0x00)
	{
		// EMPTY
	}
}
//---------------------------------------------------------------------------
void DecodeConditions(void)
{
	unsigned char chr;
	unsigned char buf[1024];
	unsigned char* ptr;
	int val1, val2, depth = 0;
	int attrptr;

	ptr = buf;

	attrptr = 0;
	while(1)
	{
		chr = databuf[curpos++];

		if(chr=='(')
		{
			SeenSaveLog("( ");
			depth++;
		}
		else if(chr==')')
		{
			SeenSaveLog(") ");
			depth--;

			if(depth <= 0)
			{
				break;
			}
		}
		else if(chr == 0x26)
		{
			SeenSaveLog("and ");
		}
		else if(chr == 0x27)
		{
			SeenSaveLog("or ");
		}
		else if(chr == 0x58)
		{
			// ȊH
			Error();
		}
		else if(chr >= 0x36 && chr <= 0x55)
		{
			// 
			val1 = ReadValue();
			val2 = ReadValue();

			switch(chr)
			{
			case 0x36:
				SeenSaveLog("!= b %d n %d ", val1, val2);
				break;

			case 0x37:
				SeenSaveLog("== b %d n %d ", val1, val2);
				break;

			case 0x38:
				SeenSaveLog("!= b %d b %d ", val1, val2);
				break;

			case 0x39:
				SeenSaveLog("== b %d b %d ", val1, val2);
				break;

			case 0x3a:
				SeenSaveLog("!= v %d n %d ", val1, val2);
				break;

			case 0x3b:
				SeenSaveLog("== v %d n %d ", val1, val2);
				break;

			case 0x41:
			case 0x42:
				SeenSaveLog("& v %d n %d ", val1, val2);
				break;

			case 0x43:
				SeenSaveLog("^ v %d n %d ", val1, val2);
				break;

			case 0x44:
				SeenSaveLog("> v %d n %d ", val1, val2);
				break;

			case 0x45:
				SeenSaveLog("< v %d n %d ", val1, val2);
				break;

			case 0x46:
				SeenSaveLog(">= v %d n %d ", val1, val2);
				break;

			case 0x47:
				SeenSaveLog("<= v %d n %d ", val1, val2);
				break;

			case 0x48:
				SeenSaveLog("!= v %d v %d ", val1, val2);
				break;

			case 0x49:
				SeenSaveLog("== v %d v %d ", val1, val2);
				break;

			case 0x4f:
			case 0x50:
				SeenSaveLog("& v %d v %d ", val1, val2);
				break;

			case 0x51:
				SeenSaveLog("^ v %d v %d ", val1, val2);
				break;

			case 0x52:
				SeenSaveLog("> v %d v %d ", val1, val2);
				break;

			case 0x53:
				SeenSaveLog("< v %d v %d ", val1, val2);
				break;

			case 0x54:
				SeenSaveLog(">= v %d v %d ", val1, val2);
				break;

			case 0x55:
				SeenSaveLog("<= v %d v %d ", val1, val2);
				break;

			default:
				*ptr++ = 0;
				break;
			}
		}
	}
}
//---------------------------------------------------------------------------
char* Flag_GetStr(int i)
{
	static char buf[100];

	if((i >= 0) && (i < 100))
	{
		sprintf(buf, "s%d", i);
		return buf;
	}

	Error();
	return NULL;
}
